# -*- coding: utf-8 -*-
"""
openSMILE LICENSE
=================
Authors: Florian Eyben, Felix Weninger, Martin Woellmer, Bjoern Schuller
Copyright (C) 2008-2013, Institute for Human-Machine Communication, TUM
Copyright (C) 2013-2014, audEERING UG (haftungsbeschrÃ¤nkt)
http://www.audeering.com/research-and-open-source/files/openSMILE-open-source-license.txt
"""
from __future__ import print_function, division, absolute_import

import os
import re
import random
import subprocess
from six import add_metaclass
from collections import Mapping
from abc import abstractproperty, ABCMeta, abstractmethod

import numpy as np

from odin.preprocessing.base import Extractor
from odin.utils import (is_string, get_script_path, ctext, is_number,
                        get_logpath, uuid)

__all__ = [
    'openSMILEloudness',
    'openSMILEf0',
    'openSMILEpitch',
    'openSMILEsad'
]

# ===========================================================================
# Helper
# ===========================================================================
def verify_dependencies():
  try:
    command = "SMILExtract -h"
    output = subprocess.check_output(command, shell=True,
                                     stderr=subprocess.STDOUT)
  except subprocess.CalledProcessError:
    raise Exception("Can't find SMILExtract executable")
  else:
    m = re.search('openSMILE version (.*)', str(output, 'utf-8'),
                  re.MULTILINE)
    if m:
      opensmile_version = m.group(1)
      print('Found openSMILE:', ctext(opensmile_version, 'magenta'))


def _get_conf_file(name):
  conf_path = os.path.join(
      os.path.dirname(__file__),
      'confs',
      name)
  if not os.path.exists(conf_path):
    raise RuntimeError("Cannot find config file at path: %s" % conf_path)
  with open(conf_path, 'r') as f:
    return f.read()

# ===========================================================================
# Base
# ===========================================================================
@add_metaclass(ABCMeta)
class _openSMILEbase(Extractor):

  def __init__(self, sr):
    verify_dependencies()
    super(_openSMILEbase, self).__init__()
    self._id = uuid(length=25)
    self.sr = sr
    self._first_config_generated = False
    self._conf = _get_conf_file('%s.cfg' % self.__class__.__name__)
    self._log_level = -1

  # ==================== abstract ==================== #
  def set_log_level(self, level):
    """ level: {int, bool}
      if `int`, log-level in integer (from 0 - 9) higher
      means more detail, -1 for turning off the log.
      if True, set the log-level to default: 2
    """
    if is_number(level):
      self._log_level = int(level)
    elif bool(level):
      self._log_level = 2
    else:
      self._log_level = -1
    return self

  @abstractproperty
  def config(self):
    pass

  @abstractmethod
  def _post_processing(self, X):
    pass

  @property
  def config_path(self):
    return os.path.join(get_logpath(), '%s%s.cfg' %
        (self.__class__.__name__, self._id))

  # ==================== utilities ==================== #
  def _update_config(self):
    if self.sr is None:
      return
    with open(self.config_path, 'w') as f:
      f.write(self._conf.format(**self.config))

  def _transform(self, X):
    # ====== file input file ====== #
    raw = None
    path = None
    if isinstance(X, Mapping):
      if 'path' in X:
        path = X['path']
      if 'sr' in X:
        if self.sr is None:
          self.sr = X['sr']
          self._update_config()
          self._first_config_generated = True
        elif self.sr != X['sr']:
          raise ValueError("Given sample rate: %d, but the audio file has "
                           "sample rate: %d" % (self.sr, X['sr']))
      if 'raw' in X:
        raw = X['raw']
    elif is_string(X):
      path = X
    elif isinstance(X, np.ndarray):
      raw = X
    else:
      raise ValueError("openSMILE extractor require path to audio file.")
    # no sample rate specified, cannot generate appropriate config
    if self.sr is None:
      raise RuntimeError("Cannot acquire sample rate for the input.")
    # ====== first time generate config ====== #
    if not self._first_config_generated:
      self._first_config_generated = True
      self._update_config()
    # ====== extract SAD ====== #
    unique_id = os.getpid() + random.randint(0, 10e8)
    inpath = os.path.join(
        get_logpath(), '%s%d.wav' % (self.__class__.__name__, unique_id))
    outpath = os.path.join(
        get_logpath(), '%s%d.csv' % (self.__class__.__name__, unique_id))
    try:
      if path is None or not os.path.exists(path):
        if raw is None:
          raise RuntimeError("openSMILE require input audio file, since "
              "we cannot find any audio file, it is required to provide "
              "raw array and sample rate, so the audio file will be cached.")
        from soundfile import write
        write(inpath, data=raw, samplerate=self.sr)
        path = inpath
      # if in debug mode or not
      command = 'SMILExtract -loglevel %d -C %s -I %s -O %s' % \
          (self._log_level, self.config_path, path, outpath)
      os.system(command)
      results = np.genfromtxt(outpath, dtype='float32',
                              delimiter=',', skip_header=0)
    except Exception as e:
      import traceback; traceback.print_exc()
      raise e
    finally:
      if os.path.exists(inpath):
        os.remove(inpath)
      if os.path.exists(outpath):
        os.remove(outpath)
    # ====== post-processing ====== #
    X_update = self._post_processing(results)
    if not isinstance(X_update, dict):
      raise ValueError("_post_processing must return a dictionary.")
    return X_update

# ===========================================================================
# Extractor
# ===========================================================================
class openSMILEf0(_openSMILEbase):
  """ HNR based on F0 harmonics and other spectral peaks """

  def __init__(self, frame_length, step_length=None,
               fmin=52, fmax=620, voicingCutoff=0.7,
               n_candidates=8, sr=None):
    super(openSMILEf0, self).__init__(sr=sr)
    self.frame_length = float(frame_length)
    if step_length is None:
      step_length = frame_length / 4
    self.step_length = float(step_length)
    self._config_file = _get_conf_file('openSMILEf0.cfg')
    self.fmin = int(fmin)
    self.fmax = int(fmax)
    self.voicingCutoff = float(voicingCutoff)
    self.n_candidates = int(n_candidates)

  @property
  def config(self):
    args = {'framesize': self.frame_length,
            'framestep': self.step_length,
            'fmin': self.fmin,
            'fmax': self.fmax,
            'ncandidates': self.n_candidates,
            'voicingCutoff': self.voicingCutoff}
    return args

  def _post_processing(self, X):
    f0 = X[:-1]
    return {'f0': f0[:, np.newaxis]}

class openSMILEloudness(_openSMILEbase):
  """ Loudness via simple auditory band model
  """

  def __init__(self, frame_length, step_length=None,
               nmel=40, fmin=20, fmax=None,
               to_intensity=False, sr=None):
    super(openSMILEloudness, self).__init__(sr=sr)
    self.frame_length = float(frame_length)
    if step_length is None:
      step_length = frame_length / 4
    self.step_length = float(step_length)
    self.nmel = int(nmel)
    self._config_file = _get_conf_file('openSMILEloudness.cfg')
    self.fmin = int(fmin)
    self.fmax = fmax
    self.to_intensity = bool(to_intensity)

  @property
  def config(self):
    if self.fmax is None:
      self.fmax = self.sr // 2
    args = {'framesize': self.frame_length,
            'framestep': self.step_length,
            'fmin': self.fmin,
            'fmax': self.fmax,
            'nmel': self.nmel}
    return args

  def _post_processing(self, X):
    name = 'loudness'
    if self.to_intensity:
      X = X * 60
      name = 'intensity'
    return {name: X[:, np.newaxis]}

class openSMILEpitch(_openSMILEbase):
  """
  Parameters
  ----------
  fmin: int
      minimum pitch frequency
  fmax: int
      maximum pitch frequency
  voicingCutoff_pitch: float [0-1]
      This sets the voicing probability threshold for pitch
      detection [0.0 - 1.0]. Frames with voicing probability
      values above this threshold will be considered as voiced.
      for 'shs', default value is `0.7`
      for 'acf', default value is `0.55`
  f0min: int
      minimum F0-frequency
  f0max: int
      maximum F0-frequency
  n_candidates: int [1-20]
      The number of F0 candidates to output [1-20] (0 disables
      ouput of candidates AND their voicing probs.).
  voicingCutoff_f0: float [0-1]
      same value applied for F0 extracting
  method: 'shs', 'acf'
      shs: subharmonic summation
      acf: autocorrelation function
  f0: bool
      if True, return F0 frequency also.
  loudness: bool
      if True, append loudness values to the output
      `L : L = (I/I0)^0.3`
      where `I` is intensity values, and I0 is constant for 60 dB and
      max. amplitude = 1.0 (I0 = 10^-6 or 0.000001)
      (for sample values normalised to the range -1..1)",0)
      intensity values is mean of squared input values
      multiplied by a Hamming window.
  voiceProb: bool
      if True, append `sap` speech activities probabilities to
      the output
  """

  def __init__(self, frame_length, step_length=None,
               window='gauss',
               fmin=52, fmax=620, voicingCutoff_pitch=0.7,
               f0min=64, f0max=400, n_candidates=8, voicingCutoff_f0=0.45,
               method='shs',
               f0=False, loudness=False, voiceProb=False,
               sr=None):
    super(openSMILEpitch, self).__init__(sr=sr)
    # ====== config ====== #
    self.frame_length = float(frame_length)
    if step_length is None:
      step_length = frame_length / 4
    self.step_length = float(step_length)
    self.window = str(window)
    # for pitch
    self.fmin = int(fmin)
    self.fmax = int(fmax)
    self.voicingCutoff_pitch = np.clip(float(voicingCutoff_pitch), 0., 1.)
    # for F0
    self.f0 = bool(f0)
    self.f0_config = _get_conf_file('smileF0.cfg')
    self.f0min = int(f0min)
    self.f0max = int(f0max)
    self.n_candidates = int(n_candidates)
    self.voicingCutoff_f0 = np.clip(float(voicingCutoff_f0), 0., 1.)
    # others
    self.loudness = bool(loudness)
    self.voiceProb = bool(voiceProb)
    # ====== method ====== #
    method = str(method).lower()
    self.method_name = method
    if method == 'shs':
      self.method = _get_conf_file('prosodyShs.cfg')
    elif method == 'acf':
      self.method = _get_conf_file('prosodyAcf.cfg')
    else:
      raise ValueError("Only two methods support: acf (autocorrelation function), "
          "and shs (subharmonic summation).")

  @property
  def config(self):
    # check method for extracting pitch
    method = self.method.format(**{
        'fmin': self.fmin, 'fmax': self.fmax,
        'voicingCutoff': self.voicingCutoff_pitch})
    method_path = self.config_path + '.method'
    with open(method_path, 'w') as f:
      f.write(method)
    # check method for extracting F0
    if self.f0:
      if self.method_name == 'acf':
        turn_on_specscale = ''
      else:
        turn_on_specscale = ';'
      f0_config = self.f0_config.format(**{
          'fmin': self.f0min, 'fmax': self.f0max,
          'nCandidates': self.n_candidates,
          'turn_on_specscale': turn_on_specscale,
          'voicingCutoff': self.voicingCutoff_f0})
      f0_path = self.config_path + '.f0'
      with open(f0_path, 'w') as f:
        f.write(f0_config)
      f0_path = '\\{%s}' % f0_path
      f0_flag = ';F0'
    else:
      f0_path = ''
      f0_flag = ''
    return {'framesize': self.frame_length, 'framestep': self.step_length,
            'window': self.window, 'method': '\\{%s}' % method_path,
            'f0': f0_path, 'f0_flag': f0_flag}

  def _post_processing(self, X):
    X = X[:, 1:] # remove timestamp
    if self.method_name == 'shs':
      X_loud = X[:, 2:3] # loud
      X_sap = X[:, 1:2] # sap
      X_pitch = X[:, 0:1] # pitch
    else:
      X_loud = X[:, 2:3] # loud
      X_sap = X[:, 0:1] # sap
      X_pitch = X[:, 1:2] # pitch
    ret = {'pitch': X_pitch}
    if self.f0:
      ret['f0'] = X[:, 3:4]
    if self.loudness:
      ret['loudness'] = X_loud
    if self.voiceProb:
      ret['sap'] = X_sap
    return ret

class openSMILEsad(_openSMILEbase):
  """ SMILEsad
  NOTE: This is only for testing, this method is really not efficience
  """

  def __init__(self, frame_length, step_length=None,
               window='ham', threshold=None, sr=None,
               output_name='sad'):
    super(openSMILEsad, self).__init__(sr=sr)
    # ====== verifying ====== #
    from odin.fuel import openSMILEsad as SADmodel
    sad_ds = SADmodel.get_dataset()
    self._lstm_path = sad_ds['lstmvad_rplp18d_12.net']
    self._init_path = sad_ds['rplp18d_norm.dat']
    # ====== config ====== #
    self.frame_length = float(frame_length)
    if step_length is None:
      step_length = frame_length / 4
    self.step_length = float(step_length)
    self.window = str(window)
    self.threshold = None if threshold is None \
        else np.clip(threshold, -1., 1.)
    self._output_name = str(output_name)

  @property
  def config(self):
    return {'framesize': self.frame_length, 'framestep': self.step_length,
            'netfile': self._lstm_path, 'initfile': self._init_path,
            'hifreq': self.sr // 2 if self.sr < 16000 else 8000,
            'window': self.window}

  def _post_processing(self, X):
    X = X[:, -1] # remove timestamp
    if is_number(self.threshold):
      X = (X >= self.threshold).astype("bool")
    return {self.output_name: X}



from odin.ml.base import evaluate
from odin.ml.decompositions import *
from odin.ml.fast_tsne import fast_tsne
from odin.ml.gmm_classifier import GMMclassifier
from odin.ml.gmm_tmat import GMM, Tmatrix
from odin.ml.ivector import Ivector
from odin.ml.plda import PLDA
from odin.ml.scoring import (Scorer, VectorNormalizer, compute_class_avg,
                             compute_wccn, compute_within_cov)



from __future__ import print_function, division, absolute_import

from abc import ABCMeta, abstractmethod, abstractproperty
from six import add_metaclass

import numpy as np

from odin.utils import ctext, is_number, one_hot
from odin.visual import (print_confusion, plot_detection_curve,
                         plot_confusion_matrix, plot_save, figure,
                         plot_Cnorm)

from sklearn.base import (BaseEstimator, TransformerMixin, DensityMixin,
                          ClassifierMixin, RegressorMixin)
from sklearn.metrics import log_loss, accuracy_score, confusion_matrix

def evaluate(y_true, y_pred_proba=None, y_pred_log_proba=None,
             labels=None, title='', path=None,
             xlims=None, ylims=None, print_log=True):
  from odin.backend import to_llr
  from odin.backend.metrics import (det_curve, compute_EER, roc_curve,
                                    compute_Cavg, compute_Cnorm,
                                    compute_minDCF)

  def format_score(s):
    return ctext('%.4f' % s if is_number(s) else s, 'yellow')
  nb_classes = None
  # ====== check y_pred ====== #
  if y_pred_proba is None and y_pred_log_proba is None:
    raise ValueError("At least one of `y_pred_proba` or `y_pred_log_proba` "
                     "must not be None")
  y_pred_llr = to_llr(y_pred_proba) if y_pred_log_proba is None \
      else to_llr(y_pred_log_proba)
  nb_classes = y_pred_llr.shape[1]
  y_pred = np.argmax(y_pred_llr, axis=-1)
  # ====== check y_true ====== #
  if isinstance(y_true, (tuple, list)):
    y_true = np.array(y_true)
  if y_true.ndim == 2: # convert one-hot to labels
    y_true = np.argmax(y_true, axis=-1)
  # ====== check labels ====== #
  if labels is None:
    labels = [str(i) for i in range(nb_classes)]
  # ====== scoring ====== #
  if y_pred_proba is None:
    ll = 'unknown'
  else:
    ll = log_loss(y_true=y_true, y_pred=y_pred_proba)
  acc = accuracy_score(y_true=y_true, y_pred=y_pred)
  cm = confusion_matrix(y_true=y_true, y_pred=y_pred)
  # C_norm
  cnorm, cnorm_arr = compute_Cnorm(y_true=y_true,
                                   y_score=y_pred_llr,
                                   Ptrue=[0.1, 0.5],
                                   probability_input=False)
  if y_pred_log_proba is not None:
    cnorm_, cnorm_arr_ = compute_Cnorm(y_true=y_true,
                                       y_score=y_pred_log_proba,
                                       Ptrue=[0.1, 0.5],
                                       probability_input=False)
    if np.mean(cnorm) > np.mean(cnorm_): # smaller is better
      cnorm, cnorm_arr = cnorm_, cnorm_arr_
  # DET
  Pfa, Pmiss = det_curve(y_true=y_true, y_score=y_pred_llr)
  eer = compute_EER(Pfa=Pfa, Pmiss=Pmiss)
  minDCF = compute_minDCF(Pfa, Pmiss)[0]
  # PRINT LOG
  if print_log:
    print(ctext("--------", 'red'), ctext(title, 'cyan'))
    print("Log loss :", format_score(ll))
    print("Accuracy :", format_score(acc))
    print("C_norm   :", format_score(np.mean(cnorm)))
    print("EER      :", format_score(eer))
    print("minDCF   :", format_score(minDCF))
    print(print_confusion(arr=cm, labels=labels))
  # ====== save report to PDF files if necessary ====== #
  if path is not None:
    if y_pred_proba is None:
      y_pred_proba = y_pred_llr
    from matplotlib import pyplot as plt
    plt.figure(figsize=(nb_classes, nb_classes + 1))
    plot_confusion_matrix(cm, labels)
    # Cavg
    plt.figure(figsize=(nb_classes + 1, 3))
    plot_Cnorm(cnorm=cnorm_arr, labels=labels, Ptrue=[0.1, 0.5],
               fontsize=14)
    # binary classification
    if nb_classes == 2 and \
    (y_pred_proba.ndim == 1 or (y_pred_proba.ndim == 2 and
                                y_pred_proba.shape[1] == 1)):
      fpr, tpr = roc_curve(y_true=y_true, y_score=y_pred_proba.ravel())
      # det curve
      plt.figure()
      plot_detection_curve(Pfa, Pmiss, curve='det',
                           xlims=xlims, ylims=ylims, linewidth=1.2)
      # roc curve
      plt.figure()
      plot_detection_curve(fpr, tpr, curve='roc')
    # multiclasses
    else:
      y_true = one_hot(y_true, nb_classes=nb_classes)
      fpr_micro, tpr_micro, _ = roc_curve(y_true=y_true.ravel(),
                                          y_score=y_pred_proba.ravel())
      Pfa_micro, Pmiss_micro = Pfa, Pmiss
      fpr, tpr = [], []
      Pfa, Pmiss = [], []
      for i, yi in enumerate(y_true.T):
        curve = roc_curve(y_true=yi, y_score=y_pred_proba[:, i])
        fpr.append(curve[0])
        tpr.append(curve[1])
        curve = det_curve(y_true=yi, y_score=y_pred_llr[:, i])
        Pfa.append(curve[0])
        Pmiss.append(curve[1])
      plt.figure()
      plot_detection_curve(fpr_micro, tpr_micro, curve='roc',
                           linewidth=1.2, title="ROC Micro")
      plt.figure()
      plot_detection_curve(fpr, tpr, curve='roc',
                           labels=labels, linewidth=1.0,
                           title="ROC for each classes")
      plt.figure()
      plot_detection_curve(Pfa_micro, Pmiss_micro, curve='det',
                           xlims=xlims, ylims=ylims, linewidth=1.2,
                           title="DET Micro")
      plt.figure()
      plot_detection_curve(Pfa, Pmiss, curve='det',
                           xlims=xlims, ylims=ylims,
                           labels=labels, linewidth=1.0,
                           title="DET for each classes")
    plot_save(path)

@add_metaclass(ABCMeta)
class Evaluable(object):
  """ Evaluable """

  @abstractproperty
  def labels(self):
    raise NotImplementedError

  def evaluate(self, X, y, labels=None, title='', path=None,
               xlims=None, ylims=None, print_log=True):
    from odin.backend import to_llr
    # ====== check inputs ====== #
    if labels is None:
      labels = self.labels
    if isinstance(y, (tuple, list)):
      y = np.array(y)
    if y.ndim == 2: # convert one-hot to labels
      y = np.argmax(y, axis=-1)
    # ====== proba ====== #
    if hasattr(self, 'predict_proba'):
      y_pred_prob = self.predict_proba(X)
    else:
      y_pred_prob = None
    # ====== log proba ====== #
    if hasattr(self, 'predict_log_proba'):
      y_pred_log_prob = self.predict_log_proba(X)
    elif y_pred_prob is not None:
      y_pred_log_prob = to_llr(y_pred_prob)
    else:
      raise ValueError('Class "%s" must has: `predict_proba` or `predict_log_proba`'
                       ' method.' % self.__class__.__name__)
    evaluate(y_true=y, y_pred_proba=y_pred_prob, y_pred_log_proba=y_pred_log_prob,
             labels=labels, title=title, path=path, xlims=xlims, ylims=ylims,
             print_log=print_log)
    return self



from __future__ import absolute_import, division, print_function

import inspect
from typing import Callable, Optional, Type, Union

import tensorflow as tf
from tensorflow.python.keras import Model, Sequential
from tensorflow.python.keras import layers as layer_module
from tensorflow.python.keras.layers import Dense, Lambda
from tensorflow_probability.python.distributions import Distribution
from tensorflow_probability.python.layers import DistributionLambda

from odin.bay.distribution_layers import VectorDeterministicLayer
from odin.bay.helpers import Statistic, kl_divergence
from odin.networks.distribution_util_layers import Moments, Sampling

__all__ = ['DenseDeterministic', 'DenseDistribution']


class DenseDeterministic(Dense):
  """ Similar to `keras.Dense` layer but return a
  `tensorflow_probability.Deterministic` distribution to represent the output,
  hence, make it compatible to probabilistic frameworks
  """

  def __init__(self,
               units,
               activation=None,
               use_bias=True,
               kernel_initializer='glorot_uniform',
               bias_initializer='zeros',
               kernel_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               **kwargs):
    super(DenseDeterministic,
          self).__init__(units=units,
                         activation=activation,
                         use_bias=use_bias,
                         kernel_initializer=kernel_initializer,
                         bias_initializer=bias_initializer,
                         kernel_regularizer=kernel_regularizer,
                         bias_regularizer=bias_regularizer,
                         activity_regularizer=activity_regularizer,
                         kernel_constraint=kernel_constraint,
                         bias_constraint=bias_constraint,
                         **kwargs)

  def call(self, inputs, **kwargs):
    outputs = super(DenseDeterministic, self).call(inputs)
    return VectorDeterministicLayer()(outputs)


class DenseDistribution(Sequential):
  """ using `Dense` layer to parameterize tensorflow_probability `Distribution`

  Parameters
  ----------
  units : `int`
    number of output units.
  posterior : {`DistributionLambda`, `callable`, `type`}
    posterior distribution, the class or a callable can be given for later
    initialization.
  prior : {`None`, `tensorflow_probability.Distribution`}
    prior distribution, used for calculating KL divergence later.
  use_bias : `bool` (default=`True`)

  call_mode : `odin.bay.helpers.Statistic` (default=`Statistic.SAMPLE`)

  name : `str` (default='DenseDistribution')

  Return
  ------
  sample, mean, variance, stddev : [n_samples, batch_size, units]
    depend on the `call_mode`, multiple statistics could be returned
  """

  def __init__(
      self,
      units,
      posterior: Union[DistributionLambda, Type[DistributionLambda], Callable],
      prior: Optional[Distribution] = None,
      activation='linear',
      use_bias=True,
      call_mode: Statistic = Statistic.DIST,
      posterior_kwargs={},
      name="DenseDistribution"):
    assert prior is None or isinstance(prior, Distribution), \
     "prior can be None or instance of tensorflow_probability.Distribution"
    assert isinstance(call_mode, Statistic), \
      "call_mode must be instance of odin.bay.helpers.Statistic"
    units = int(units)
    use_bias = bool(use_bias)
    # process the posterior
    if isinstance(posterior, DistributionLambda):
      layer = posterior
    elif (isinstance(posterior, type) and
          issubclass(posterior, DistributionLambda)) or \
          isinstance(posterior, Callable):
      args = inspect.getfullargspec(posterior).args
      posterior_kwargs = {
          i: j for i, j in posterior_kwargs.items() if i in args
      }
      layer = posterior(units, **posterior_kwargs)
    else:
      raise ValueError("No support for posterior of type: %s" %
                       str(type(layer)))
    # layer must be DistributionLambda
    assert isinstance(layer, DistributionLambda), \
      "The callable must return instance of DistributionLambda, but given: %s" \
        % (str(type(layer)))
    # create layers
    params_size = layer.params_size(units)
    layers = [
        Dense(params_size, activation=activation, use_bias=use_bias), layer
    ]
    super(DenseDistribution, self).__init__(layers=layers, name=name)
    # basics
    self._units = units
    self._params_size = params_size
    self._posterior = posterior
    self._prior = prior
    self._activation = activation
    self._use_bias = bool(use_bias)
    self._call_mode = call_mode
    self._last_distribution = None
    self._posterior_kwargs = posterior_kwargs
    # check class init
    for arg in inspect.getfullargspec(self.__init__).args:
      if arg not in locals():
        raise ValueError("Invalid subclassing %s, the arguments to __init__ "
                         "must contain DenseDistribution arguments." %
                         str(self.__class__))

  def get_config(self):
    config = {
        'name': self.name,
        'configs': {
            'units': self._units,
            'posterior': self._posterior,
            'prior': self._prior,
            'activation': self._activation,
            'use_bias': self._use_bias,
            'call_mode': self._call_mode,
            'posterior_kwargs': self._posterior_kwargs
        }
    }
    if self._build_input_shape:
      config['build_input_shape'] = self._build_input_shape
    return config

  @classmethod
  def from_config(cls, config, custom_objects=None):
    if 'name' in config:
      name = config['name']
      build_input_shape = config.get('build_input_shape')
    else:
      name = None
      build_input_shape = None
    kw = {
        i: config['configs'][i]
        for i in inspect.getfullargspec(cls.__init__).args
        if i in config['configs']
    }
    model = cls(name=name, **kw)
    if not model.inputs and build_input_shape:
      model.build(build_input_shape)
    return model

  @property
  def prior(self):
    return self._prior

  @property
  def distribution_layer(self):
    return self.layers[-1]

  @property
  def posterior(self):
    """ Return the last parametrized distribution, i.e. the result from `call`
    """
    return self._last_distribution

  @property
  def units(self):
    return self._units

  def _apply_distribution(self, x):
    if hasattr(x, '_distribution') and \
      x._distribution == self._last_distribution:
      dist = x._distribution
    else:
      dist = super(DenseDistribution, self).call(x)
    return dist

  def mean(self, x):
    dist = self._apply_distribution(x)
    y = Moments(mean=True, variance=False)(dist)
    setattr(y, '_distribution', dist)
    self._last_distribution = y._distribution
    return y

  def variance(self, x):
    dist = self._apply_distribution(x)
    y = Moments(mean=False, variance=True)(dist)
    setattr(y, '_distribution', dist)
    self._last_distribution = y._distribution
    return y

  def stddev(self, x):
    v = self.variance(x)
    y = Lambda(tf.math.sqrt)(v)
    setattr(y, '_distribution', v._distribution)
    return y

  def sample(self, x, n_samples=1):
    if n_samples is None or n_samples <= 0:
      n_samples = 1
    dist = self._apply_distribution(x)
    y = Sampling(n_samples=n_samples)(dist)
    setattr(y, '_distribution', dist)
    self._last_distribution = y._distribution
    return y

  def call(self, x, training=None, n_samples=1, mode=None):
    """
    Parameters
    ----------
    x : {`numpy.ndarray`, `tensorflow.Tensor`}

    training : {`None`, `bool`} (default=`None`)

    n_samples : {`None`, `int`} (default=`1`)

    mode : {`None`, `odin.bay.helpers.Statistic`} (default=`None`)
      decide which of the statistics will be return from the distribution,
      this value will overide the default value of the class
    """
    results = []
    variance = None
    call_mode = self._call_mode if not isinstance(mode, Statistic) else mode
    # special case only need the distribution
    if Statistic.DIST == call_mode:
      dist = super(DenseDistribution, self).call(x)
      self._last_distribution = dist
      return dist
    # convert to tensor modes
    if Statistic.SAMPLE in call_mode:
      results.append(self.sample(x, n_samples=n_samples))
    if Statistic.MEAN in call_mode:
      results.append(self.mean(x))
    if Statistic.VAR in call_mode or Statistic.STDDEV in call_mode:
      variance = self.variance(x)
      if Statistic.VAR in call_mode:
        results.append(variance)
      if Statistic.STDDEV in call_mode:
        y = Lambda(tf.math.sqrt)(variance)
        setattr(y, '_distribution', variance._distribution)
        results.append(y)
    if Statistic.DIST in call_mode:
      assert len(results) > 0
      results.append(results[0]._distribution)
    return results[0] if len(results) == 1 else tuple(results)

  def kl_divergence(self, x=None, prior=None, analytic_kl=True, n_samples=1):
    """
    Parameters
    ---------
    x : `Tensor`
      optional input for parametrizing the distribution, if not given,
      used the last result from `call`

    prior : instance of `tensorflow_probability.Distribution`
      prior distribution of the latent

    analytic_kl : `bool` (default=`True`)
      using closed form solution for calculating divergence,
      otherwise, sampling with MCMC

    n_samples : `int` (default=`1`)
      number of MCMC sample if `analytic_kl=False`

    Return
    ------
    kullback_divergence : Tensor [n_samples, batch_size, ...]
    """
    if n_samples is None:
      n_samples = 1
    if prior is None:
      prior = self._prior
    elif self._prior is None:
      self._prior = prior
    assert isinstance(prior, Distribution), "prior is not given!"

    if x is not None:
      self(x, mode=Statistic.DIST)
    if self.posterior is None:
      raise RuntimeError(
          "DenseDistribution must be called to create the distribution before "
          "calculating the kl-divergence.")

    kullback_div = kl_divergence(q=self.posterior,
                                 p=prior,
                                 use_analytic_kl=bool(analytic_kl),
                                 q_sample=int(n_samples),
                                 auto_remove_independent=True)
    if analytic_kl:
      kullback_div = tf.expand_dims(kullback_div, axis=0)
      if n_samples > 1:
        ndims = kullback_div.shape.ndims
        kullback_div = tf.tile(kullback_div, [n_samples] + [1] * (ndims - 1))
    return kullback_div

  def log_prob(self, x):
    """ Calculating the log probability (i.e. log likelihood) """
    assert self.units == x.shape[-1], \
      "Number of features mismatch, units=%d  input_shape=%s" % \
        (self.units, str(x.shape))
    if self.posterior is None:
      raise RuntimeError(
          "DenseDistribution must be called to create the distribution before "
          "calculating the log-likelihood.")
    return self.posterior.log_prob(x)

  def __repr__(self):
    return self.__str__()

  def __str__(self):
    return '<DenseDistribution units:%d #params:%g posterior:%s prior:%s>' %\
      (self.units, self._params_size / self.units,
       self.layers[-1].__class__.__name__,
       self.prior.__class__.__name__)



from __future__ import absolute_import, division, print_function

import collections

import numpy as np
import tensorflow as tf
from sklearn.mixture import GaussianMixture
from tensorflow.python import keras
from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras.layers import Dense
from tensorflow_probability.python import bijectors as tfb
from tensorflow_probability.python import distributions as tfd
from tensorflow_probability.python.layers.distribution_layer import (
    DistributionLambda, _get_convert_to_tensor_fn, _serialize,
    _serialize_function)
from tensorflow_probability.python.layers.internal import \
    distribution_tensor_coercible as dtc
from tensorflow_probability.python.layers.internal import \
    tensor_tuple as tensor_tuple

__all__ = ['MixtureDensityNetwork']

_COV_TYPES = ('none', 'diag', 'full', 'tril')


class MixtureDensityNetwork(Dense):
  """A mixture of Gaussian Keras layer.

  Parameters
  ----------
  units : `int`
    number of output features for each component.
  n_components : `int` (default=`2`)
    The number of mixture components.
  covariance_type : {'none', 'diag', 'full', 'tril'}
    String describing the type of covariance parameters to use.
    Must be one of:
      'none' (each component has its own single variance).
      'diag' (each component has its own diagonal covariance matrix),
      'tril' (lower triangle matrix),
      'full' (each component has its own general covariance matrix),

  """

  def __init__(self,
               units,
               n_components=2,
               covariance_type='none',
               convert_to_tensor_fn=tfd.Distribution.sample,
               softplus_scale=True,
               validate_args=False,
               activation='linear',
               use_bias=True,
               kernel_initializer='glorot_uniform',
               bias_initializer='zeros',
               kernel_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               **kwargs):
    covariance_type = str(covariance_type).lower()
    assert covariance_type in _COV_TYPES, \
    "No support for covariance_type: '%s', the support value are: %s" % \
      (covariance_type, ', '.join(_COV_TYPES))
    self._covariance_type = covariance_type
    self._n_components = int(n_components)
    self._validate_args = bool(validate_args)
    self._convert_to_tensor_fn = _get_convert_to_tensor_fn(convert_to_tensor_fn)
    self._softplus_scale = bool(softplus_scale)
    # We'll need to keep track of who's calling who since the functional
    # API has a different way of injecting `_keras_history` than the
    # `keras.Sequential` way.
    self._enter_dunder_call = False
    # ====== calculating the number of parameters ====== #
    if covariance_type == 'none':
      component_params_size = 2 * units
    elif covariance_type == 'diag':  # only the diagonal
      component_params_size = units + units
    elif covariance_type == 'tril':  # lower triangle
      component_params_size = units + units * (units + 1) // 2
    elif covariance_type == 'full':  # full matrix
      component_params_size = units + units * units
    else:
      raise NotImplementedError
    self._component_params_size = component_params_size
    params_size = self.n_components + self.n_components * component_params_size
    self._event_size = units

    super(MixtureDensityNetwork,
          self).__init__(units=params_size,
                         activation=activation,
                         use_bias=use_bias,
                         kernel_initializer=kernel_initializer,
                         bias_initializer=bias_initializer,
                         kernel_regularizer=kernel_regularizer,
                         bias_regularizer=bias_regularizer,
                         activity_regularizer=activity_regularizer,
                         kernel_constraint=kernel_constraint,
                         bias_constraint=bias_constraint,
                         **kwargs)

  @property
  def event_size(self):
    return self._event_size

  @property
  def covariance_type(self):
    return self._covariance_type

  @property
  def n_components(self):
    return self._n_components

  @property
  def component_params_size(self):
    return self._component_params_size

  def __call__(self, inputs, *args, **kwargs):
    self._enter_dunder_call = True
    distribution, _ = super(MixtureDensityNetwork,
                            self).__call__(inputs, *args, **kwargs)
    self._enter_dunder_call = False
    return distribution

  def call(self, inputs, *args, **kwargs):
    dense_kwargs = dict(kwargs)
    dense_kwargs.pop('training', None)
    params = super(MixtureDensityNetwork, self).call(inputs, *args,
                                                     **dense_kwargs)
    n_components = tf.convert_to_tensor(value=self.n_components,
                                        name='n_components',
                                        dtype_hint=tf.int32)
    # ====== mixture weights ====== #
    mixture_coefficients = params[..., :n_components]
    mixture_dist = tfd.Categorical(logits=mixture_coefficients,
                                   validate_args=self._validate_args,
                                   name="MixtureWeights")
    # ====== initialize the components ====== #
    params = tf.reshape(
        params[..., n_components:],
        tf.concat([tf.shape(input=params)[:-1], [n_components, -1]], axis=0))
    if bool(self._softplus_scale):
      scale_fn = lambda x: tf.math.softplus(x) + tfd.softplus_inverse(1.0)
    else:
      scale_fn = lambda x: x

    if self.covariance_type == 'none':
      cov = 'IndependentNormal'
      loc_params, scale_params = tf.split(params, 2, axis=-1)
      scale_params = scale_params
      components_dist = tfd.Independent(tfd.Normal(
          loc=loc_params,
          scale=scale_fn(scale_params),
          validate_args=self._validate_args),
                                        reinterpreted_batch_ndims=1)
    #
    elif self.covariance_type == 'diag':
      cov = 'MultivariateNormalDiag'
      loc_params, scale_params = tf.split(params, 2, axis=-1)
      components_dist = tfd.MultivariateNormalDiag(
          loc=loc_params,
          scale_diag=scale_fn(scale_params),
          validate_args=self._validate_args)
    #
    elif self.covariance_type == 'tril':
      cov = 'MultivariateNormalTriL'
      loc_params = params[..., :self.event_size]
      scale_params = scale_fn(params[..., self.event_size:])
      scale_tril = tfb.ScaleTriL(diag_shift=np.array(
          1e-5, params.dtype.as_numpy_dtype()),
                                 validate_args=self._validate_args)
      components_dist = tfd.MultivariateNormalTriL(
          loc=loc_params,
          scale_tril=scale_tril(scale_params),
          validate_args=self._validate_args)
    #
    elif self.covariance_type == 'full':
      cov = 'MultivariateNormalFull'
      loc_params = params[..., :self.event_size]
      scale_params = tf.reshape(
          scale_fn(params[..., self.event_size:]),
          tf.concat(
              [tf.shape(input=params)[:-1], (self.event_size, self.event_size)],
              axis=0))
      components_dist = tfd.MultivariateNormalFullCovariance(
          loc=loc_params,
          covariance_matrix=scale_params,
          validate_args=self._validate_args)
    else:
      raise NotImplementedError
    # ====== finally the mixture ====== #
    d = tfd.MixtureSameFamily(mixture_distribution=mixture_dist,
                              components_distribution=components_dist,
                              validate_args=False,
                              name="Mixture%s" % cov)
    # Wraps the distribution to return both dist and concrete value."""
    value_is_seq = isinstance(d.dtype, collections.Sequence)
    maybe_composite_convert_to_tensor_fn = (
        (lambda d: tensor_tuple.TensorTuple(self._convert_to_tensor_fn(d)))
        if value_is_seq else self._convert_to_tensor_fn)
    distribution = dtc._TensorCoercible(  # pylint: disable=protected-access
        distribution=d,
        convert_to_tensor_fn=maybe_composite_convert_to_tensor_fn)
    value = distribution._value()  # pylint: disable=protected-access
    value._tfp_distribution = distribution  # pylint: disable=protected-access
    if value_is_seq:
      value.shape = value[-1].shape
      value.get_shape = value[-1].get_shape
      value.dtype = value[-1].dtype
      distribution.shape = value[-1].shape
      distribution.get_shape = value[-1].get_shape
    else:
      distribution.shape = value.shape
      distribution.get_shape = value.get_shape
    if self._enter_dunder_call:
      # Its critical to return both distribution and concretization
      # so Keras can inject `_keras_history` to both. This is what enables
      # either to be used as an input to another Keras `Model`.
      return distribution, value
    return distribution

  def compute_output_shape(self, input_shape):
    input_shape = tensor_shape.TensorShape(input_shape)
    input_shape = input_shape.with_rank_at_least(2)
    if tensor_shape.dimension_value(input_shape[-1]) is None:
      raise ValueError(
          'The innermost dimension of input_shape must be defined, but saw: %s'
          % input_shape)
    # the number of output units is equal to event_size, not number of
    # hidden units
    return input_shape[:-1].concatenate(self.event_size)

  def get_config(self):
    """Returns the config of this layer. """
    config = {
        'convert_to_tensor_fn': _serialize(self._convert_to_tensor_fn),
        'covariance_type': self._covariance_type,
        'validate_args': self._validate_args,
        'n_components': self._n_components,
        'softplus_scale': self._softplus_scale,
    }
    base_config = super(MixtureDensityNetwork, self).get_config()
    base_config.update(config)
    return base_config



from __future__ import print_function, division, absolute_import

import os
import base64
import struct
import pickle
import hashlib
import zipfile
from io import BytesIO
from six import string_types

import numpy as np
import scipy as sp

# ===========================================================================
# Helper
# ===========================================================================
def to_password(password, salt=None):
  """ This function return the byte data of hashed password """
  if isinstance(password, string_types):
    password = password.encode('utf-8')
  password += str(salt).encode('utf-8')
  return hashlib.sha256(password).digest()

def _data_to_iobuffer(file_or_data):
  own_file = True
  # ====== string object ====== #
  if isinstance(file_or_data, string_types):
    # path to file
    if os.path.exists(file_or_data):
      filesize = os.path.getsize(file_or_data)
      infile = open(file_or_data, 'rb')
    # string object
    else:
      filesize = len(file_or_data)
      infile = BytesIO(file_or_data.encode())
  # ======  bytes object ====== #
  elif isinstance(file_or_data, bytes):
    filesize = len(file_or_data)
    infile = BytesIO(file_or_data)
  # ====== opened file ====== #
  elif hasattr(file_or_data, 'read'):
    filesize = os.fstat(file_or_data.fileno()).st_size
    infile = file_or_data
    own_file = False
  # ====== error ====== #
  else:
    raise ValueError("No support for MD5 of input type: %s" % str(file_or_data))
  return infile, filesize, own_file

# ===========================================================================
# Hashing
# ===========================================================================
def md5_checksum(file_or_path, chunksize=512 * 1024):
  """ Calculating MD5 checksum

  Parameters
  ---------
  file_or_path : object
    One of the following
      - File object
      - string path to a folder
      - string path to a file
      - Bytes array
      - Numpy array
      - List or iterator of numpy array
  chunksize : int (in bytes)
    size of each chunk for updating MD5 value
  """
  chunksize = int(chunksize)
  hash_md5 = hashlib.md5()
  own_file = False
  # ====== numpy array or list of numpy array ====== #
  if isinstance(file_or_path, np.ndarray) or \
  (isinstance(file_or_path, (tuple, list)) and
   all(isinstance(i, np.ndarray) for i in file_or_path)):
    if not isinstance(file_or_path, (tuple, list)):
      file_or_path = (file_or_path,)
    f = BytesIO()
    for arr in file_or_path:
      np.save(file=f, arr=arr, allow_pickle=False)
    f.seek(0)
    own_file = True
  # ======  path to file or folder ====== #
  elif isinstance(file_or_path, string_types):
    # a file
    if os.path.isfile(file_or_path):
      f = open(file_or_path, 'rb')
      own_file = True
    # a folder (then read all files in order)
    elif os.path.isdir(file_or_path):
      for fpath in sorted(os.listdir(file_or_path)):
        fpath = os.path.join(file_or_path, fpath)
        with open(fpath, 'rb') as f:
          for chunk in iter(lambda: f.read(chunksize), b""):
            hash_md5.update(chunk)
      return hash_md5.hexdigest()
    # just string or text
    else:
      hash_md5.update(file_or_path.encode('utf-8'))
      return hash_md5.hexdigest()
  # bytes object directly
  elif isinstance(file_or_path, bytes):
    hash_md5.update(file_or_path)
    return hash_md5.hexdigest()
  # file object or buffer object
  elif hasattr(file_or_path, 'read'):
    f = file_or_path
  # special case big custom array with shape attribute
  elif hasattr(file_or_path, 'shape'):
    if isinstance(file_or_path, sp.sparse.coo_matrix):
      file_or_path = file_or_path.todense()
    itemsize = np.dtype(file_or_path.dtype).itemsize
    batch_size = int(max(chunksize // (itemsize * np.prod(file_or_path.shape[1:])), 8))
    indices = range(0, file_or_path.shape[0] + batch_size, batch_size)
    for start, end in zip(indices, indices[1:]):
      x = file_or_path[start:end].tobytes()
      hash_md5.update(x)
    return hash_md5.hexdigest()
  # NO support
  else:
    raise ValueError("MD5 checksum has NO support for input: %s" % str(file_or_path))
  # ====== iterate over file ====== #
  for chunk in iter(lambda: f.read(chunksize), b""):
    hash_md5.update(chunk)
  if own_file:
    f.close()
  return hash_md5.hexdigest()

# ===========================================================================
# Encryption
# ===========================================================================
def encrypt_aes(file_or_data, password=None, outfile=None, iv=None, salt=None,
                mode=None, base64encode=False, chunksize=512 * 1024):
  """ Flexible implementaiton of AES encryption

  Parameters
  ----------
  file_or_data : {BufferObject, string, bytes}
    input data will be converted to bytes sequence for encryption
  password : {str, None}
    if None, a prompt will ask for inputing password
  outfile : {None, path, file}
    if None, return raw encrypted data
  iv : initial vector
    16 bytes
  salt : {None, string, bytes}
    salt for password Hashing
  mode : Cipher.AES.MODE_*
    default `None` is converted to `Crypto.Cipher.AES.MODE_CBC`
  chunksize : int
    encryption chunk, multiple of 16.
  """
  try:
    from Crypto.Cipher import AES
  except ImportError as e:
    raise ImportError("Require 'pycrypto' to run this function")
  if mode is None:
    mode = AES.MODE_CBC

  if password is None:
    password = input("Your password: ")
  assert len(password) > 0, "Password length must be greater than 0"
  password = to_password(password, salt=salt)
  # Initialization vector
  if iv is None:
    iv = os.urandom(16)
  encryptor = AES.new(password, mode, IV=iv)
  # ====== check read stream ====== #
  infile, filesize, own_file = _data_to_iobuffer(file_or_data)
  # ====== check out stream ====== #
  close_file = False
  if isinstance(outfile, string_types) and \
  os.path.exists(os.path.dirname(outfile)):
    outfile = open(str(outfile), 'wb')
    close_file = True
  elif hasattr(outfile, 'write') and hasattr(outfile, 'flush'):
    close_file = True
  else:
    outfile = BytesIO()
  # ====== some header information ====== #
  outfile.write(struct.pack('<Q', filesize))
  outfile.write(iv)
  while True:
    chunk = infile.read(chunksize)
    if bool(base64encode):
      chunk = base64.encodebytes(chunk)
    # EOF
    if len(chunk) == 0:
      break
    # doing padding to match chunk size
    elif len(chunk) % 16 != 0:
      chunk += b' ' * (16 - len(chunk) % 16)
    outfile.write(encryptor.encrypt(chunk))
  # ====== clean and return ====== #
  if own_file:
    infile.close()
  outfile.flush()
  if close_file:
    outfile.close()
  else:
    outfile.seek(0)
    data = outfile.read()
    outfile.close()
    return data

def decrypt_aes(file_or_data, password=None, outfile=None, salt=None,
                mode=None, base64encode=False, chunksize=512 * 1024):
  """ Flexible implementaiton of AES decryption

  Parameters
  ----------
  file_or_data : {BufferObject, string, bytes}
    input data will be converted to bytes sequence for encryption
  password : {str, None}
    if None, a prompt will ask for inputing password
  outfile : {None, path, file}
    if None, return raw encrypted data
  salt : {None, string, bytes}
    salt for password Hashing
  mode : Cipher.AES.MODE_*
  chunksize : int
    encryption chunk, multiple of 16.
  """
  try:
    from Crypto.Cipher import AES
  except ImportError as e:
    raise ImportError("Require 'pycrypto' to run this function")
  if mode is None:
    mode = AES.MODE_CBC

  if password is None:
    password = input("Your password: ")
  assert len(password) > 0, "Password length must be greater than 0"
  password = to_password(password, salt)
  # ====== read header ====== #
  infile, filesize, own_file = _data_to_iobuffer(file_or_data)
  origsize = struct.unpack('<Q', infile.read(struct.calcsize('Q')))[0]
  iv = infile.read(16)
  decryptor = AES.new(password, mode=AES.MODE_CBC, IV=iv)
  # ====== outfile ====== #
  close_file = False
  if isinstance(outfile, string_types) and os.path.exists(os.path.dirname(outfile)):
    outfile = open(str(outfile), 'wb')
    close_file = True
  elif hasattr(outfile, 'write') and hasattr(outfile, 'flush'):
    close_file = True
  else:
    outfile = BytesIO()
  # ====== decryption ====== #
  while True:
    chunk = infile.read(chunksize)
    if len(chunk) == 0:
      break
    chunk = decryptor.decrypt(chunk)
    if bool(base64encode):
      chunk = base64.decodebytes(chunk)
    outfile.write(chunk)
  outfile.truncate(origsize)
  # ====== clean and return ====== #
  if own_file:
    infile.close()
  outfile.flush()
  if close_file:
    outfile.close()
  else:
    outfile.seek(0)
    data = outfile.read()
    outfile.close()
    return data

# ===========================================================================
# Zip
# ===========================================================================
def zip_aes(in_path, out_path, password=None, compression=True,
            verbose=False):
  """
  Parameters
  ----------
  in_path : string
    path to a folder
  out_path : string
    path to output zip file
  """
  if password is None:
    password = input("Your password:")
  password = str(password)
  assert len(password) > 0, "`password`=%s length must be greater than 0" % password
  # ====== prepare input ====== #
  from odin.utils import get_all_files
  if not os.path.isdir(in_path):
    raise ValueError("`in_path` to %s is not a folder" % str(in_path))
  all_files = get_all_files(in_path)
  # ====== prepare output ====== #
  if not isinstance(out_path, string_types):
    raise ValueError("`out_path` must be string")
  f = zipfile.ZipFile(out_path, 'w',
        compression=zipfile.ZIP_DEFLATED if bool(compression) else
                    zipfile.ZIP_STORED,
        allowZip64=True)
  # ====== compression ====== #
  md5_map = {}
  for path in all_files:
    name = os.path.basename(path)
    md5_map[name] = md5_checksum(path)
    f.writestr(name, encrypt_aes(path, password + name))
    if verbose:
      print('Compressed: "%s"' % name, "(MD5:%s)" % md5_map[name])
  f.writestr('_MD5_CHECKSUM_', pickle.dumps(md5_map))
  f.close()

def unzip_aes(in_path, out_path=None, password=None, compression=True,
              verbose=False):
  """
  Parameters
  ----------
  in_path : string
    path to input zip file
  out_path : {string, None}
    if None, return iteration of tuple (name, decompressed data),
    otherwise, saving the decompressed data to the given folder

  """
  if password is None:
    password = input("Your password:")
  password = str(password)
  assert len(password) > 0, "`password`=%s length must be greater than 0" % password
  # ====== check in_path ====== #
  if not os.path.isfile(in_path):
    raise ValueError("`in_path` to %s is not a file" % str(in_path))
  # ====== prepare output ====== #
  if out_path is not None:
    if os.path.isfile(out_path):
      raise ValueError("`out_path` must be a folder")
    elif not os.path.exists(out_path):
      os.mkdir(out_path)
  # ====== decompress ====== #
  with zipfile.ZipFile(in_path, 'r',
          compression=zipfile.ZIP_DEFLATED if bool(compression) else
                      zipfile.ZIP_STORED,
          allowZip64=True) as fzip:
    md5_map = pickle.loads(fzip.read(name='_MD5_CHECKSUM_'))
    for name in fzip.namelist():
      if '_MD5_CHECKSUM_' == name:
        continue
      data = fzip.read(name=name)
      data = decrypt_aes(data, password=password + name)
      md5 = md5_checksum(data)
      assert md5 == md5_map[name], \
      "MD5 mismatch for data name: '%s', probably caused by wrong password!" % name
      # verbose
      if verbose:
        print('Decompressed: "%s"' % name, "(MD5:%s)" % md5)
      # save file to disk
      if out_path is not None:
        path = out_path
        for d in name.split('/')[:-1]:
          path += '/' + d
          if os.path.exists(path):
            os.mkdir(path)
        with open(os.path.join(path, name.split('/')[-1]), mode='wb') as f:
          f.write(data)
      # return iteration
      else:
        yield name, data



# -*- coding: utf-8 -*-
from __future__ import print_function

import argparse
import contextlib
import inspect
import math
import os
import platform
import re
import shutil
import signal
import subprocess
import sys
import tarfile
import tempfile
# import uuid
import time
import timeit
import types
from collections import Iterable, Iterator, Mapping, OrderedDict, deque
from itertools import chain, islice, tee
from multiprocessing import Lock, cpu_count, current_process

import numpy
from six.moves.urllib.error import HTTPError, URLError
from six.moves.urllib.request import urlopen

from odin.utils import crypto, decorators, mpi
from odin.utils.cache_utils import *
from odin.utils.crypto import md5_checksum
from odin.utils.mpi import MPI, SharedCounter, async, async_mpi, segment_list
from odin.utils.np_utils import *
from odin.utils.path_utils import *
from odin.utils.progbar import Progbar, add_notification
from odin.utils.python_utils import *

try:
  from numba import jit, autojit, vectorize, guvectorize
except:
  pass



def array_size(arr):
  """ Return size of an numpy.ndarray in bytes """
  return np.prod(arr.shape) * arr.dtype.itemsize

# ===========================================================================
# Pretty print
# ===========================================================================
def ctext(s, color='red'):
  """ Colored text support
  * BLACK
  * RED
  * GREEN
  * YELLOW
  * BLUE
  * MAGENTA
  * CYAN
  * WHITE
  * RESET
  * LIGHTBLACK_EX
  * LIGHTRED_EX
  * LIGHTGREEN_EX
  * LIGHTYELLOW_EX
  * LIGHTBLUE_EX
  * LIGHTMAGENTA_EX
  * LIGHTCYAN_EX
  * LIGHTWHITE_EX

  Note
  ----
  * colored text is longer in length and can overlength the screen
  (i.e. make the progress-bar ugly).
  * ctext(x + y + z) is BETTER than ctext(x) + ctext(y) + ctext(z)
  because it save more space.
  """
  s = str(s)
  try:
    from colorama import Fore
    color = str(color).strip()
    color = color.upper()
    # special fix for light color
    if 'LIGHT' == color[:5] and '_EX' != color[-3:]:
      color = color + '_EX'
    color = getattr(Fore, color, '')
    return color + s + Fore.RESET
  except ImportError:
    pass
  return s


def _type_name(x):
  if isinstance(x, np.dtype):
    return x.name
  return type(x).__name__


def dummy_formatter(x):
  s = str(x)
  if len(s) < 120 and '\n' not in s:
    return s
  if isinstance(x, (tuple, list)):
    return "(list)length=%d;type=%s" % \
        (len(x), _type_name(x[0]) if len(x) > 0 else '*empty*')
  if isinstance(x, np.ndarray):
    return "(ndarray)shape=%s;dtype=%s" % (str(x.shape), str(x.dtype))
  if isinstance(x, Mapping):
    return "(map)length=%d;dtype=%s" % (len(x),
        ';'.join([_type_name(i) for i in next(iter(x.items()))]))
  # dataset type
  if 'odin.fuel.dataset.Dataset' in str(type(x)):
    return ("(ds)path:%s" % x.path)
  # NNOp types
  if any('odin.nnet.base.NNOp' in str(i) for i in type.mro(type(x))):
    return "\n" + '\n'.join(['\t' + line for line in str(x).split('\n')])
  if is_string(x):
    return str(x) if len(x) < 250 else '(str)length:%d' % len(x)
  return str(x)


@contextlib.contextmanager
def UnitTimer(factor=1, name=None):
  start = timeit.default_timer()
  yield None
  end = timeit.default_timer()
  # set name for timing task
  if name is None:
    name = 'Task'
  print('"%s"' % ctext(name, 'yellow'),
      "Time:",
      ctext((end - start) / factor, 'cyan'),
      '(sec)')

# ===========================================================================
# Basics
# ===========================================================================
def is_same_shape(shape1, shape2):
  """
  Return
  ------
  True if two objects is the same shape tuple,
  otherwise, False
  """
  # if is string, evaluate to python object
  if is_string(shape1):
    shape1 = eval(shape1)
  if is_string(shape2):
    shape2 = eval(shape2)
  # tensorflow TensorShape
  if hasattr(shape1, 'as_list'):
    shape1 = shape1.as_list()
  if hasattr(shape2, 'as_list'):
    shape2 = shape2.as_list()
  # if is number, convert to shape tuple
  if is_number(shape1) or shape1 is None:
    shape1 = (shape1,)
  shape1 = as_tuple(shape1)
  if is_number(shape2) or shape2 is None:
    shape2 = (shape2,)
  shape2 = as_tuple(shape2)
  # check the number of dimension
  if len(shape1) != len(shape2):
    return False
  for s1, s2 in zip(shape1, shape2):
    if s1 is not None and s2 is not None and s1 != s2:
      return False
  return True

def type_path(obj):
  clazz = type(obj)
  return clazz.__module__ + "." + clazz.__name__

def iter_chunk(it, n):
  """ Chunking an iterator into small chunk of size `n`
  Note: this can be used to slice data into mini batches
  """
  if not isinstance(it, Iterator):
    it = iter(it)
  obj = list(islice(it, n))
  while obj:
    yield obj
    obj = list(islice(it, n))

def batching(batch_size, n=None, start=0, end=None, seed=None):
  """
  Parameters
  ----------
  batch_size: int
      number of samples for 1 single batch.
  n: {int, None}
      number of samples
  start : int (default: 0)
      starting point of the iteration (in number of sample)
  end : {int, None}
      ending point of the iteration (in number of sample),
      end = total_amount_of_sample if None
  seed : {int, None}
      random seed for shuffling the returned batches,
      no shuffling performed if seed is None

  Return
  ------
  iteration: [(start, end), (start, end), ...]
  """
  if end is None and n is None:
    raise ValueError('you must provide either `end` or `n`')
  if end is None:
    end = n
  start = int(start)
  end = int(end)
  assert end > start, "`end` must > `start`"
  batch_size = int(batch_size)
  # ====== no shuffling ====== #
  if seed is None:
    return ((i, min(i + batch_size, end))
            for i in range(start, end + batch_size, batch_size)
            if i < end)
  batches = list(range(start, end + batch_size, batch_size))
  rand = np.random.RandomState(seed)
  rand.shuffle(batches)
  return (((i, min(i + batch_size, end)))
          for i in batches
          if i < end)

def read_lines(file_path):
  if not os.path.exists(file_path):
    raise ValueError('File at path: %s does not exist.' % file_path)
  if os.path.isdir(file_path):
    raise ValueError("Path to %s is a folder" % file_path)
  lines = []
  with open(file_path, 'r') as f:
    for i in f:
      lines.append(i[:-1] if i[-1] == '\n' else i)
  return lines

# ===========================================================================
# Others
# ===========================================================================
def raise_return(e):
  raise e

_CURRENT_STDIO = None

class _LogWrapper():

  def __init__(self, stream, own_buffer):
    self.stream = stream
    self.own_buffer = bool(own_buffer)

  def write(self, message):
    # no backtrack for writing to file
    self.stream.write(message.replace("\b", ''))
    sys.__stdout__.write(message)

  def flush(self):
    self.stream.flush()
    sys.__stdout__.flush()

  def end(self):
    self.stream.flush()
    sys.__stdout__.flush()
    if self.own_buffer:
      try:
        self.stream.close()
      except Exception:
        pass

def get_stdio_path():
  return _CURRENT_STDIO

def stdio(path=None, suppress=False, stderr=True):
  """
  Parameters
  ----------
  path: {None, str, io.StringIO}
      if str or StringIO, specified path for saving all stdout (and stderr)
      if None, reset to the system default stdout and stderr
  suppress: boolean
      totally turn-off all stdout (and stdeer)
  stderr:
      apply output file with stderr also

  NOTE
  ----
  Redirect the system logging can slightly decrease the
  performance in logging intensive application.
  """
  # turn off stdio
  if suppress:
    f = open(os.devnull, "w")
    path = None
  # reset
  elif path is None:
    f = None
  # redirect to a file
  elif is_string(path):
    f = _LogWrapper(open(path, "w"), own_buffer=True)
  # redirect to a buffer
  elif is_fileobj(path):
    f = _LogWrapper(path, own_buffer=False)
    if hasattr(path, 'name'):
      path = f.name
    else:
      path = str(path)
  else:
    raise ValueError("Unsupport for path=`%s` in `stdio`." %
        str(type(path)))
  # ====== set current stdio path ====== #
  global _CURRENT_STDIO
  _CURRENT_STDIO = path
  # ====== assign stdio ====== #
  if f is None: # reset
    if isinstance(sys.stdout, _LogWrapper):
      sys.stdout.end()
    if isinstance(sys.stderr, _LogWrapper):
      sys.stderr.end()
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
  else: # redirect to file
    sys.stdout = f
    if stderr:
      sys.stderr = f

def auto_logging(log_dir=None, prefix='', num_max=None):
  """
  Parameters
  ----------
  log_dir : str
    path to directory stored all the logs
  prefix : str
    prefix to all saved log (using script name by default)
  num_max : {None, int}
    maximum number of log will be stored
  """
  if log_dir is None:
    log_dir = get_script_path()
  if not os.path.isdir(log_dir):
    raise ValueError("'%s' is not a directory" % str(log_dir))
  prefix = get_script_name() if prefix is None or len(prefix) == 0 else str(prefix)
  date_time = get_formatted_datetime(only_number=False)
  path = os.path.join(log_dir, prefix + '[%s]' % date_time + '.txt')
  # ====== check maximum number of log ====== #
  if num_max is not None:
    from stat import ST_CTIME
    num_max = int(num_max)
    past_logs = [i for i in os.listdir(log_dir) if prefix in i]
    past_logs = sorted(past_logs,
                       key=lambda x: os.stat(os.path.join(log_dir, x))[ST_CTIME],
                       reverse=True)
    # remove previous logs
    for i, name in enumerate(past_logs):
      if i >= num_max - 1:
        os.remove(os.path.join(log_dir, name))
  return stdio(path=path, suppress=False, stderr=True)

def eprint(text):
  """Print ERROR message to stderr"""
  print(ctext('[Error]', 'red') + str(text), file=sys.stderr)

def wprint(text):
  """Print WARNING message to stderr"""
  print(ctext('[Warning]', 'yellow') + str(text), file=sys.stderr)

# ===========================================================================
# Universal ID
# ===========================================================================
_uuid_chars = list(chain(map(chr, range(65, 91)),  # ABCD
                         map(chr, range(97, 123)),  # abcd
                         map(chr, range(48, 57)))) # 0123
_uuid_random_state = numpy.random.RandomState(int(str(int(time.time() * 100))[3:]))


def uuid(length=8):
  """ Generate random UUID 8 characters with very very low collision """
  # m = time.time()
  # uniqid = '%8x%4x' % (int(m), (m - int(m)) * 1000000)
  # uniqid = str(uuid.uuid4())[:8]
  uniquid = ''.join(
      _uuid_random_state.choice(_uuid_chars,
                                size=length,
                                replace=True))
  return uniquid


@contextlib.contextmanager
def change_recursion_limit(limit):
  """Temporarily changes the recursion limit."""
  old_limit = sys.getrecursionlimit()
  if old_limit < limit:
    sys.setrecursionlimit(limit)
  yield
  sys.setrecursionlimit(old_limit)


@contextlib.contextmanager
def signal_handling(sigint=None, sigtstp=None, sigquit=None):
  # We cannot handle SIGTERM, because it prevent subproces from
  # .terminate()
  orig_int = signal.getsignal(signal.SIGINT)
  orig_tstp = signal.getsignal(signal.SIGTSTP)
  orig_quit = signal.getsignal(signal.SIGQUIT)

  if sigint is not None: signal.signal(signal.SIGINT, sigint)
  if sigtstp is not None: signal.signal(signal.SIGTSTP, sigtstp)
  if sigquit is not None: signal.signal(signal.SIGQUIT, sigquit)

  yield
  # reset
  signal.signal(signal.SIGINT, orig_int)
  signal.signal(signal.SIGTSTP, orig_tstp)
  signal.signal(signal.SIGQUIT, orig_quit)


# ===========================================================================
# UniqueHasher
# ===========================================================================
class UniqueHasher(object):
  """ This hash create strictly unique hash value by using
  its memory to remember which key has been assigned

  Note
  ----
  This function use deterministic hash, which give the same
  id for all labels, whenever you call it
  """

  def __init__(self, nb_labels=None):
    super(UniqueHasher, self).__init__()
    self.nb_labels = nb_labels
    self._memory = {} # map: key -> hash_key
    self._current_hash = {} # map: hash_key -> key

  def hash(self, value):
    key = abs(hash(value))
    if self.nb_labels is not None:
      key = key % self.nb_labels
    # already processed hash
    if value in self._current_hash:
      return self._current_hash[value]
    # not yet processed
    if key in self._memory:
      if self.nb_labels is not None and \
          len(self._memory) >= self.nb_labels:
        raise Exception('All %d labels have been assigned, outbound value:"%s"' %
                        (self.nb_labels, value))
      else:
        while key in self._memory:
          key += 1
          if self.nb_labels is not None and key >= self.nb_labels:
            key = 0
    # key not in memory
    self._current_hash[value] = key
    self._memory[key] = value
    return key

  def __call__(self, value):
    return self.hash(value)

  def map(self, order, array):
    """ Re-order an ndarray to new column order """
    order = as_tuple(order)
    # get current order
    curr_order = self._current_hash.items()
    curr_order.sort(key=lambda x: x[1])
    curr_order = [i[0] for i in curr_order]
    # new column order
    order = [curr_order.index(i) for i in order]
    return array[:, order]


class FuncDesc(object):
  """ This class store the description of arguments given a function.
  Automatically match the argument and keyword-arguments in following
  order:
   - Match all positional arguments.
   - if the function has `varargs`, keep the remains positional arguments.
   - if the function has `keywords`, keep the unknown argument name in `kwargs`
   - Add the missing default kwargs.
   - Return new tuple of (arg, kwargs)
  """

  def __init__(self, func):
    super(FuncDesc, self).__init__()
    # copy information from other FuncDesc
    if isinstance(func, FuncDesc):
      self._args = func._args
      self._defaults = func._defaults
      self._is_include_args = func._inc_args
      self._is_include_kwargs = func._inc_kwargs
      self._func = func._func
    # extract information from a function or method
    elif inspect.isfunction(func) or inspect.ismethod(func) or \
    isinstance(func, decorators.functionable):
      if isinstance(func, decorators.functionable):
        sign = inspect.signature(func.function)
      else:
        sign = inspect.signature(func)
      self._args = [n for n, p in sign.parameters.items()
                    if p.kind not in (inspect.Parameter.VAR_POSITIONAL,
                                      inspect.Parameter.VAR_KEYWORD)]
      self._is_include_args = any(i.kind == inspect.Parameter.VAR_POSITIONAL
                           for i in sign.parameters.values())
      self._is_include_kwargs = any(i.kind == inspect.Parameter.VAR_KEYWORD
                             for i in sign.parameters.values())
      self._defaults = {n: p.default
                        for n, p in sign.parameters.items()
                        if p.default != inspect.Parameter.empty}
      self._func = func
    else:
      raise ValueError("`func` must be function, method, or FuncDesc, "
                       "but given: %s" % str(type(func)))
    self.__name__ = self._func.__name__

  def __getstate__(self):
    import dill
    # check if need to re-load function module during setstate
    if not is_lambda(self._func):
      module = inspect.getmodule(self._func)
      func_module_key = [i
                         for i, j in list(sys.modules.items())
                         if j == module][0]
      func_module_name = module.__name__
      func_module_path = inspect.getfile(module)
    else:
      func_module_key = None
      func_module_name = None
      func_module_path = None
    # using dill to dump function
    func_str = dill.dumps(self._func)
    return (func_str,
            func_module_key, func_module_name, func_module_path,
            self.__name__,
            self._args, self._is_include_args,
            self._is_include_kwargs, self._defaults)

  def __setstate__(self, states):
    import dill
    (func_str,
     func_module_key, func_module_name, func_module_path,
     self.__name__,
     self._args, self._is_include_args,
     self._is_include_kwargs, self._defaults) = states
    # iterate through all loaded modules to find module contain given function
    if func_module_key is not None and \
    func_module_name is not None and \
    func_module_path is not None:

      found = None
      for name, module in list(sys.modules.items()):
        try:
          if inspect.getfile(module) == func_module_path or \
          (module.__name__.split('.')[-1] == func_module_name and
           self.__name__ in dir(module)):
            found = (name, module)
        except Exception as e:
          pass

      if found is None:
        # TODO: search for module in sys.path
        import imp
        path = func_module_path
        script_path = get_script_path()
        folders = func_module_path.split('/')
        for i in range(1, len(folders)):
          if os.path.exists(path) and os.path.isfile(path):
            break
          path = os.path.join(script_path, '/'.join(folders[-i:]))
        imp.load_source(name=func_module_name, pathname=path)
      elif found[0] != func_module_name:
        sys.modules[func_module_name] = sys.modules[found[0]]

    # load function
    self._func = dill.loads(func_str)

  @property
  def name(self):
    return self.__name__

  @property
  def qualname(self):
    """ The qualified name of the class, function, method, descriptor,
    or generator instance.

    New in version 3.3.
    """
    assert hasattr(self._func, '__qualname__'), \
    "The attribute `__qualname__` only be introduced from python 3.3"
    return self._func.__qualname__

  @property
  def path(self):
    """ Return the absolute path to the script contain the function """
    return inspect.getfile(inspect.getmodule(self._func))

  @property
  def args(self):
    return tuple(self._args)

  @property
  def defaults(self):
    return dict(self._defaults)

  @property
  def is_args(self):
    return self._is_include_args

  @property
  def is_kwargs(self):
    return self._is_include_kwargs

  def _match(self, *args, **kwargs):
    keywords = OrderedDict()
    for name, val in zip(self.args, args):
      keywords[name] = val
    # extra args
    if self._is_include_args and len(self.args) < len(args):
      args = args[len(self.args):]
    else:
      args = ()
    # remove extra kwargs in inc_kwargs=False
    if not self._is_include_kwargs:
      kwargs = {name: kwargs[name] for name in self._args
                if name in kwargs}
    keywords.update(kwargs)
    # ====== update the default ====== #
    for name, val in self._defaults.items():
      if name not in keywords:
        keywords[name] = val
    return args, keywords

  def __call__(self, *args, **kwargs):
    args, kwargs = self._match(*args, **kwargs)
    return self._func(*args, **kwargs)

  def __str__(self):
    s = "<%s>args:%s defaults:%s varargs:%s keywords:%s" % \
        (ctext(self._func.__name__, 'cyan'), self._args, self._defaults,
            self._is_include_args, self._is_include_kwargs)
    return s

# ===========================================================================
# ArgCOntrol
# ===========================================================================
def args_parse(descriptions):
  """ Shortcut for parsing the argument from terminal
  command, by a list of following tuple:
    [(name, help, enum_values, default_value), ...]

  Parameter
  ---------
  descriptions : tuple, list
    a list of tuple contain 1 to 3 of following information
    (name, help, enum_values, default_value), where:
    - name (string) is the name of argument can be `name` positional,
      `-name` optional, or `--name` flag trigger
    - help (string) is text for printing help
    - enum_values (tuple, list, None) list of string (or Number) contain
      acceptable values (set to None if no values given)
    - default_value (Number, string) default value for the argument

  Note
  ----
  The order is different from `ArgController.add`
  """
  args = ArgController()
  for desc in descriptions:
    if len(desc) == 1:
      args.add(name=desc[0], help='')
    elif len(desc) == 2:
      args.add(name=desc[0], help=desc[1])
    elif len(desc) == 3:
      args.add(name=desc[0], help=desc[1], enum=desc[2])
    elif len(desc) == 4:
      args.add(name=desc[0], help=desc[1], enum=desc[2], default=desc[3])
    else:
      raise ValueError("No support for given description: %s" % str(desc))
  return args.parse()

class ArgController(object):
  """ Simple interface to argparse """

  def __init__(self, print_parsed=True):
    super(ArgController, self).__init__()
    self.parser = None
    self._require_input = False
    self.args_preprocessor = {}
    self.args_enum = {}
    self.name = []
    self.print_parsed = print_parsed

  def _process_name(self, name):
    if '--' == name[:2]:
      name = name[2:]
    elif '-' == name[0]:
      name = name[1:]
    name = name.replace('-', '_')
    return name

  def _is_positional(self, name):
    if name[0] != '-' and name[:2] != '--':
      return True
    return False

  def _parse_input(self, key, val):
    key = self._process_name(key)
    # ====== search if manual preprocessing available ====== #
    for i, preprocess in self.args_preprocessor.items():
      if key == i and preprocess is not None:
        return preprocess(val)
    # ====== auto preprocess ====== #
    try:
      val = float(val)
      if int(val) == val:
        val = int(val)
    except Exception as e:
      val = str(val)
    return val

  def add(self, name, help, default=None, enum=None,
          preprocess=None):
    """ NOTE: if the default value is not given, the argument is
    required

    Parameters
    ----------
    name: str
        [-name] for optional parameters
        [name] for positional parameters
    help: str
        description of the argument
    default: {str, Number}
        default value for the argument
    enum : {tuple, list}
        list or tuple of acceptant values, for post-process
        validating
    preprocess: call-able
        a function to perform preprocessing of the parsed argument
    """
    if self.parser is None:
      self.parser = argparse.ArgumentParser(
          description='Automatic argument parser (yes,true > True; no,false > False)',
          add_help=True)
    # ====== NO default value ====== #
    if default is None:
      if self._is_positional(name):
        self.parser.add_argument(name, help=help, type=str, action="store",
            metavar='')
      else:
        self.parser.add_argument(name, help=help, type=str, action="store",
            required=True, metavar='')
      self._require_input = True
    # ====== boolean default value ====== #
    elif isinstance(default, bool):
      help += ' (default: %s)' % str(default)
      self.parser.add_argument(name, help=help,
                               action="store_%s" % str(not default).lower())
      preprocess = lambda x: bool(x)
    # ====== add defaults value ====== #
    else:
      help += ' (default: %s)' % str(default)
      self.parser.add_argument(name, help=help, type=str, action="store",
          default=str(default), metavar='')

    # ====== store preprocess dictionary ====== #
    name = self._process_name(name)
    if not hasattr(preprocess, '__call__'):
      preprocess = None
    self.args_preprocessor[name] = preprocess
    # store the enum values
    if enum is not None and not isinstance(enum, (tuple, list)):
      enum = (enum,)
    self.args_enum[name] = enum
    return self

  def parse(self, config_path=None):
    if self.parser is None:
      raise Exception('Call add to assign at least 1 argument for '
                      'for the function.')
    exit_now = False
    # ====== validate arguments ====== #
    try:
      if len(sys.argv) == 1 and self._require_input:
        self.parser.print_help()
        exit_now = True
      else:
        args = self.parser.parse_args()
    except Exception as e:
      # if specfy version or help, don't need to print anything else
      if all(i not in ['-h', '--help', '-v', '--version']
             for i in sys.argv):
        self.parser.print_help()
      exit_now = True
    if exit_now:
      exit()
    # ====== parse the arguments ====== #
    try:
      args = {i: self._parse_input(i, j)
              for i, j in args._get_kwargs()}
    except Exception as e:
      print('Error parsing argument with name "%s"' % str(e))
      self.parser.print_help()
      exit()
    # ====== checking enumerate values ====== #
    for name, val in args.items():
      enum = self.args_enum[self._process_name(name)]
      if enum is None:
        continue
      if val not in enum:
        raise ValueError("Argument with name '%s' given value: `%s`, but "
                         "only accept one of the following: %s" %
                         (name, val, enum))
    # ====== reset everything ====== #
    self.parser = None
    self.args_preprocessor = {}
    self.args_enum = {}
    self._require_input = False
    # ====== print parsed values ====== #
    if self.print_parsed:
      max_len = max(len(i) for i in args.keys())
      max_len = '%-' + str(max_len) + 's'
      print('\n******** Parsed arguments ********')
      for name, val in sorted(args.items()):
        print(max_len % name, ': ', val)
      print('**********************************\n')
    # convert it to struct
    return struct(args)

# ===========================================================================
# Simple math and processing
# ===========================================================================
def as_tuple_of_shape(x):
  if not isinstance(x, (tuple, list)):
    x = (x,)
  if is_number(x[0]):
    x = (x,)
  return x

def axis_normalize(axis, ndim,
                   return_tuple=False):
  """ Normalize the axis
   * -1 => ndim - 1
   * 10 => 10 % ndim
   * None => list(range(ndim))

  Parameters
  ----------
  return_tuple: bool
      if True, always return a tuple of normalized axis
  """
  if isinstance(axis, np.ndarray):
    axis = axis.ravel().tolist()
  if ndim == 0:
    return ()
  if axis is None or \
  (isinstance(axis, (tuple, list)) and len(axis) == 1 and axis[0] is None):
    return list(range(ndim))
  if not hasattr(axis, '__len__'):
    axis = int(axis) % ndim
    return (axis,) if return_tuple else axis
  axis = as_tuple(axis, t=int)
  return tuple([i % ndim for i in axis])


def flatten_list(x, level=None):
  """
  Parameters
  ----------
  level: int, or None
      how deep the function go into element of x to search for list and
      flatten it. If None is given, flatten all list found.

  Example
  -------
  >>> l = [1, 2, 3, [4], [[5], [6]], [[7], [[8], [9]]]]
  >>> print(flatten_list(l, level=1))
  >>> # [1, 2, 3, 4, [5], [6], [7], [[8], [9]]]
  >>> print(flatten_list(l, level=2))
  >>> # [1, 2, 3, 4, 5, 6, 7, [8], [9]]
  >>> print(flatten_list(l, level=None))
  >>> # [1, 2, 3, 4, 5, 6, 7, 8, 9]
  """
  if isinstance(x, Iterator):
    x = list(x)
  if level is None:
    level = 10e8
  if not isinstance(x, (tuple, list)):
    return [x]
  if any(isinstance(i, (tuple, list)) for i in x):
    _ = []
    for i in x:
      if isinstance(i, (tuple, list)) and level > 0:
        _ += flatten_list(i, level - 1)
      else:
        _.append(i)
    return _
  return x


# ===========================================================================
# Python
# ===========================================================================
class struct(dict):

  '''Flexible object can be assigned any attribtues'''

  def __init__(self, *args, **kwargs):
    super(struct, self).__init__(*args, **kwargs)
    # copy all dict to attr
    for i, j in self.items():
      if is_string(i) and not hasattr(self, i):
        super(struct, self).__setattr__(i, j)

  def __setattr__(self, name, val):
    super(struct, self).__setattr__(name, val)
    super(struct, self).__setitem__(name, val)

  def __setitem__(self, x, y):
    super(struct, self).__setitem__(x, y)
    if is_string(x):
      super(struct, self).__setattr__(x, y)


class bidict(dict):
  """ Bi-directional dictionary (i.e. a <-> b)
  Note
  ----
  When you iterate over this dictionary, it will be a doubled size
  dictionary
  """

  def __init__(self, *args, **kwargs):
    super(bidict, self).__init__(*args, **kwargs)
    # this is duplication
    self._inv = dict()
    for i, j in self.items():
      self._inv[j] = i

  @property
  def inv(self):
    return self._inv

  def __setitem__(self, key, value):
    super(bidict, self).__setitem__(key, value)
    self._inv[value] = key
    return None

  def __getitem__(self, key):
    if key not in self:
      return self._inv[key]
    return super(bidict, self).__getitem__(key)

  def update(self, *args, **kwargs):
    for k, v in dict(*args, **kwargs).items():
      self[k] = v
      self._inv[v] = k

  def __delitem__(self, key):
    del self._inv[super(bidict, self).__getitem__(key)]
    return dict.__delitem__(self, key)


# Under Python 2, 'urlretrieve' relies on FancyURLopener from legacy
# urllib module, known to have issues with proxy management
if sys.version_info[0] == 2:
  def urlretrieve(url, filename, reporthook=None, data=None):
    '''
    This function is adpated from: https://github.com/fchollet/keras
    Original work Copyright (c) 2014-2015 keras contributors
    '''
    def chunk_read(response, chunk_size=8192, reporthook=None):
      total_size = response.info().get('Content-Length').strip()
      total_size = int(total_size)
      count = 0
      while 1:
        chunk = response.read(chunk_size)
        if not chunk:
          break
        count += 1
        if reporthook:
          reporthook(count, chunk_size, total_size)
        yield chunk

    response = urlopen(url, data)
    with open(filename, 'wb') as fd:
      for chunk in chunk_read(response, reporthook=reporthook):
        fd.write(chunk)
else:
  from six.moves.urllib.request import urlretrieve

def get_file(fname, origin, outdir, verbose=False):
  '''
  Parameters
  ----------
  fname: output file name
  origin: url, link
  outdir: path to output dir
  '''
  fpath = os.path.join(outdir, fname)
  # ====== remove empty folder ====== #
  if os.path.exists(fpath):
    if os.path.isdir(fpath) and len(os.listdir(fpath)) == 0:
      shutil.rmtree(fpath)
  # ====== download package ====== #
  if not os.path.exists(fpath):
    if verbose:
      prog = Progbar(target=-1,
                     name="Downloading: %s" % os.path.basename(origin),
                     print_report=True, print_summary=True)

    def dl_progress(count, block_size, total_size):
      if verbose:
        if prog.target < 0:
          prog.target = total_size
        else:
          prog.add(count * block_size - prog.seen_so_far)
    ###
    error_msg = 'URL fetch failure on {}: {} -- {}'
    try:
      try:
        urlretrieve(origin, fpath, dl_progress)
      except URLError as e:
        raise Exception(error_msg.format(origin, e.errno, e.reason))
      except HTTPError as e:
        raise Exception(error_msg.format(origin, e.code, e.msg))
    except (Exception, KeyboardInterrupt) as e:
      if os.path.exists(fpath):
        os.remove(fpath)
      raise
  return fpath

# ===========================================================================
# Python utilities
# ===========================================================================
def get_all_files(path, filter_func=None):
  ''' Recurrsively get all files in the given path '''
  file_list = []
  if os.access(path, os.R_OK):
    for p in os.listdir(path):
      p = os.path.join(path, p)
      if os.path.isdir(p):
        file_list += get_all_files(p, filter_func)
      else:
        if filter_func is not None and not filter_func(p):
          continue
        # remove dump files of Mac
        if '.DS_Store' in p or '.DS_STORE' in p or \
            '._' == os.path.basename(p)[:2]:
          continue
        file_list.append(p)
  return file_list


def get_all_ext(path):
  """ Recurrsively get all extension of files in the given path

  Parameters
  ----------
  path : str
    input folder

  """
  file_list = []
  if os.access(path, os.R_OK):
    for p in os.listdir(path):
      p = os.path.join(path, p)
      if os.path.isdir(p):
        file_list += get_all_ext(p)
      else:
        # remove dump files of Mac
        if '.DS_Store' in p or '.DS_STORE' in p or \
            '._' == os.path.basename(p)[:2]:
          continue
        ext = p.split('.')
        if len(ext) > 1:
          file_list.append(ext[-1])
  file_list = list(set(file_list))
  return file_list


def folder2bin(path):
  """ This function read all files within a Folder
  in binary mode,
  then, store all the data in a dictionary mapping:
  `relative_path -> binary_data`
  """
  if not os.path.isdir(path):
    raise ValueError('`path`=%s must be a directory.' % path)
  path = os.path.abspath(path)
  files = get_all_files(path)
  data = {}
  for f in files:
    name = f.replace(path + '/', '')
    with open(f, 'rb') as f:
      data[name] = f.read()
  return data


def bin2folder(data, path, override=False):
  """ Convert serialized data from `folder2bin` back
  to a folder at `path`

  Parameters
  ----------
  data: {string, dict}
      if string, `data` can be pickled string, or path to a file.
      if dict, `data` is the output from `folder2bin`
  path: string
      path to a folder
  override: bool
      if True, override exist folder at `path`
  """
  # ====== check input ====== #
  if is_string(data):
    if os.path.isfile(data):
      with open(data, 'rb') as f:
        data = pickle.load(f)
    else:
      data = pickle.loads(data)
  if not isinstance(data, dict):
    raise ValueError("`data` must be dictionary type, or string, or path to file.")
  # ====== check outpath ====== #
  path = os.path.abspath(str(path))
  if not os.path.exists(path):
    os.mkdir(path)
  elif os.path.isfile(path):
    raise ValueError("`path` must be path to a directory.")
  elif os.path.isdir(path):
    if not override:
      raise RuntimeError("Folder at path:%s exist, cannot override." % path)
    shutil.rmtree(path)
    os.mkdir(path)
  # ====== deserialize ====== #
  for name, dat in data.items():
    with open(os.path.join(path, name), 'wb') as f:
      f.write(dat)
  return path


# ===========================================================================
# Package utils
# ===========================================================================
def package_installed(name, version=None):
  import pip
  for i in pip.get_installed_distributions():
    if name.lower() == i.key.lower() and \
    (version is None or version == i.version):
      return True
  return False


def package_list(include_version=False):
  """
  Return
  ------
  ['odin', 'lasagne', 'keras', ...] if include_version is False
  else ['odin==8.12', 'lasagne==25.18', ...]
  """

  all_packages = []
  import pip
  for i in pip.get_installed_distributions():
    all_packages.append(i.key +
        (('==' + i.version) if include_version is True else ''))
  return all_packages


def get_module_from_path(identifier, path='.',
                         prefix='', suffix='', exclude='',
                         prefer_compiled=False, return_error=False):
  ''' Algorithms:
   - Search all files in the `path` matched `prefix` and `suffix`
   - Exclude all files contain any str in `exclude`
   - Sorted all files based on alphabet
   - Load all modules based on `prefer_compiled`
   - return list of identifier found in all modules

  Parameters
  ----------
  identifier : str
      identifier of object, function or anything in script files

  prefix : str
      prefix of file to search in the `path`

  suffix : str
      suffix of file to search in the `path`

  path : str
      searching path of script files

  exclude : str, list(str)
      any files contain str in this list will be excluded

  prefer_compiled : bool
      True mean prefer .pyc file, otherwise prefer .py

  return_error : bool (default: False)
      return all the error happened during loading the modules

  Returns
  -------
  list(object, function, ..) :
      any thing match given identifier in all found script file
  errors (optional) :
      error while loading any of the script

  Notes
  -----
  File with multiple . character my procedure wrong results

  If the script run this this function match the searching process, a
  infinite loop may happen!

  * This function try to import each modules and find desire function,
  it may mess up something.

  '''
  import re
  import imp
  # ====== validate input ====== #
  if exclude == '':
    exclude = []
  if type(exclude) not in (list, tuple, numpy.ndarray):
    exclude = [exclude]
  prefer_flag = 1 if prefer_compiled else -1
  # ====== create pattern and load files ====== #
  pattern = re.compile(r"^%s.*%s\.pyc?" % (prefix, suffix)) # py or pyc
  file_name = os.listdir(path)
  file_name = [f for f in file_name
           if pattern.match(f) and
           sum([i in f for i in exclude]) == 0]
  # ====== remove duplicated pyc files ====== #
  file_name = sorted(file_name,
                     key=lambda x: prefer_flag * len(x)) # pyc is longer
  # .pyc go first get overrided by .py
  file_name = sorted({f.split('.')[0]: f for f in file_name}.values())
  # ====== load all modules ====== #
  modules = []
  modules_error = {}
  # NOTE: this will load the module ignore the relative import path
  # For example: for module A.B.C
  # `from A.B import C` will result 'A.B.C' -> C
  # `imp.load_source`, here, will result 'C' -> C
  for fname in file_name:
    try:
      if '.pyc' in fname:
        modules.append(
            imp.load_compiled(fname.split('.')[0],
                              os.path.join(path, fname))
        )
      else:
        modules.append(
            imp.load_source(fname.split('.')[0],
                            os.path.join(path, fname))
        )
    except Exception as e:
      modules_error[fname] = str(e)
      eprint("Cannot loading modules from file: '%s' - %s" %
        (ctext(fname, 'yellow'), ctext(str(e), 'red')))
  # ====== Find all identifier in modules ====== #
  ids = []
  for m in modules:
    for i in inspect.getmembers(m):
      if identifier in i:
        ids.append(i[1])
  # remove duplicate py
  if bool(return_error):
    return ids, modules_error
  return ids


def ordered_set(seq):
  seen = {}
  result = []
  for marker in seq:
    if marker in seen: continue
    seen[marker] = 1
    result.append(marker)
  return result


def dict_union(*dicts, **kwargs):
  r"""Return union of a sequence of disjoint dictionaries.

  Parameters
  ----------
  dicts : dicts
      A set of dictionaries with no keys in common. If the first
      dictionary in the sequence is an instance of `OrderedDict`, the
      result will be OrderedDict.
  \*\*kwargs
      Keywords and values to add to the resulting dictionary.

  Raises
  ------
  ValueError
      If a key appears twice in the dictionaries or keyword arguments.

  """
  dicts = list(dicts)
  if dicts and isinstance(dicts[0], OrderedDict):
    result = OrderedDict()
  else:
    result = {}
  for d in list(dicts) + [kwargs]:
    duplicate_keys = set(result.keys()) & set(d.keys())
    if duplicate_keys:
      raise ValueError("The following keys have duplicate entries: {}"
                       .format(", ".join(str(key) for key in
                                         duplicate_keys)))
    result.update(d)
  return result


# ===========================================================================
# PATH, path manager
# ===========================================================================
@contextlib.contextmanager
def TemporaryDirectory(add_to_path=False):
  """
  add_to_path: bool
      temporary add the directory to system $PATH
  """
  temp_dir = tempfile.mkdtemp()
  if add_to_path:
    os.environ['PATH'] = temp_dir + ':' + os.environ['PATH']
  current_dir = os.getcwd()
  os.chdir(temp_dir)
  yield temp_dir
  os.chdir(current_dir)
  if add_to_path:
    os.environ['PATH'] = os.environ['PATH'].replace(temp_dir + ':', '')
  shutil.rmtree(temp_dir)


def get_tempdir():
  return tempfile.mkdtemp()


def _get_managed_path(folder, name, override, is_folder=False,
                      root='~', odin_base=True):
  if root == '~':
    root = os.path.expanduser('~')
  datadir_base = os.path.join(root, '.odin') if odin_base else root
  if not os.path.exists(datadir_base):
    os.mkdir(datadir_base)
  elif not os.access(datadir_base, os.W_OK):
    raise Exception('Cannot acesss path: ' + datadir_base)
  datadir = os.path.join(datadir_base, folder)
  if not os.path.exists(datadir):
    os.makedirs(datadir)
  # ====== check given path with name ====== #
  if is_string(name):
    datadir = os.path.join(datadir, str(name))
    if os.path.exists(datadir) and override:
      if os.path.isfile(datadir): # remove file
        os.remove(datadir)
      else: # remove and create new folder
        shutil.rmtree(datadir)
    if is_folder and not os.path.exists(datadir):
      os.mkdir(datadir)
  return datadir


def get_datasetpath(name=None, override=False, is_folder=True, root='~'):
  return _get_managed_path('datasets', name, override,
                           is_folder=is_folder, root=root)


def get_figpath(name=None, override=False, root='~'):
  return _get_managed_path('figs', name, override,
                           is_folder=True, root=root)

def get_modelpath(name=None, override=False, root='~'):
  """ Default model path for saving O.D.I.N networks """
  return _get_managed_path('models', name, override,
                           is_folder=True, root=root)


def get_logpath(name=None, override=False, increasing=True,
                odin_base=True, root='~'):
  """
  name : (string) name of log file
  override : (bool) if True, override old log file if exist
  increasing : (bool) append a number (e.g. "log.0.txt"...), if log file exists
  odin_base : (bool) if True, create log file under '.odin/logs' folder
  root : (string) root folder that contain all the log file
  """
  def _check_logpath(log_path):
    main_path, ext = os.path.splitext(log_path)
    main_path = main_path.split('.')
    try:
      current_log_index = int(main_path[-1])
      main_path = main_path[:-1]
    except ValueError:
      current_log_index = 0
    main_path = '.'.join(main_path)
    # ====== increase log index until found a new file ====== #
    while True:
      path = main_path + '.' + str(current_log_index) + ext
      if not os.path.exists(path):
        break
      current_log_index += 1
    return main_path + '.' + str(current_log_index) + ext

  path = _get_managed_path('logs' if odin_base else '', name, override,
                           is_folder=False, root=root,
                           odin_base=odin_base)
  if increasing:
    path = _check_logpath(path)
  return path


def get_exppath(tag, name=None, override=False, prompt=False,
                root='~'):
  """ Specific path for experiments results

  Parameters
  ----------
  tag: string
      specific tag for the task you are working on
  name: string
      name of the folder contains all the results (NOTE: the
      name can has subfolder)
  override: bool
      if True, remove exist folder
  prompt: bool
      if True, display prompt and require (Y) input before
      delete old folder, if (N), the program exit.
  root: string
      root path for the results (default: "~/.odin")
  """
  path = _get_managed_path('exp', tag, False,
                           is_folder=True, root=root,
                           odin_base=False)
  # only return the main folder
  if name is None:
    pass
  # return the sub-folder
  else:
    name = str(name).split('/')
    for i in name:
      path = os.path.join(path, i)
      if not os.path.exists(path):
        os.mkdir(path)
  # ====== check if override ====== #
  if override and len(os.listdir(path)) > 0:
    if prompt:
      user_cmd = raw_input('Do you want to delete "%s" (Y for yes):').lower()
      if user_cmd != 'y':
        exit()
    shutil.rmtree(path)
    os.mkdir(path)
  return path

# ===========================================================================
# Misc
# ===========================================================================
def run_script(s, wait=True, path='/tmp'):
  """
  Parameters
  ----------
  s : string
    python script

  wait : bool (default: True)
    blocking the current process until finishing the script

  path : string (default: '/tmp')
    saving the script to a temporary path before running

  Return
  ------
  status: of executed command, if `wait`=True else return Popen object
  out: (string, utf-8) - stdout message
  err: (string, utf-8) - stderr message
  """
  # ====== path preprocessing ====== #
  if path is None:
    path = '/tmp'
  if os.path.isdir(path):
    path = os.path.join(path, 'tmp_%s' % uuid(length=25) + '.py')
  path = os.path.abspath(path)
  # ====== script preprocessing ====== #
  s = str(s).strip()
  # ====== run the script ====== #
  try:
    with open(path, 'w') as f:
      f.write(s)
    cmd = subprocess.Popen("python %s" % path, shell=True,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out, err) = cmd.communicate()
    if wait:
      status = cmd.wait()
    else:
      status = cmd
  finally:
    if os.path.exists(path):
      os.remove(path)
  return status, str(out, encoding='utf-8'), str(err, encoding='utf-8')

def exec_commands(cmds, print_progress=True):
  ''' Execute a command or list of commands in parallel with multiple process
  (as much as we have CPU)

  Parameters
  ----------
  cmds: str or list of str
      string represent command you want to run

  Return
  ------
  failed: list of failed command

  '''
  if not cmds: return [] # empty list
  if not isinstance(cmds, (list, tuple)):
    cmds = [cmds]
  else:
    cmds = list(cmds)

  def done(p):
    return p.poll() is not None

  def success(p):
    return p.returncode == 0

  max_task = cpu_count()
  processes = []
  processes_map = {}
  failed = []
  if print_progress:
    prog = Progbar(target=len(cmds), name="Execute Commands")
  while True:
    while cmds and len(processes) < max_task:
      task = cmds.pop()
      p = subprocess.Popen(task, shell=True)
      processes.append(p)
      processes_map[p] = task
      # print process
      if print_progress:
        prog.add(1)

    for p in processes:
      if done(p):
        if success(p):
          processes.remove(p)
        else:
          failed.append(processes_map[p])

    if not processes and not cmds:
      break
    else:
      time.sleep(0.005)
  return failed


def save_wav(path, s, fs):
  from scipy.io import wavfile
  wavfile.write(path, fs, s)


def play_audio(data, fs, volumn=1, speed=1):
  ''' Play audio from numpy array.

  Parameters
  ----------
  data : numpy.ndarray
          signal data
  fs : int
          sample rate
  volumn: float
          between 0 and 1
  speed: float
          > 1 mean faster, < 1 mean slower

  Note
  ----
  Only support play audio on MacOS
  '''
  import soundfile as sf
  import os

  data = numpy.asarray(data, dtype=numpy.int16)
  if data.ndim == 1:
    channels = 1
  else:
    channels = data.shape[1]
  with TemporaryDirectory() as temppath:
    path = os.path.join(temppath, 'tmp_play.wav')
    with sf.SoundFile(path, 'w', fs, channels, subtype=None,
        endian=None, format=None, closefd=None) as f:
      f.write(data)
    os.system('afplay -v %f -q 1 -r %f %s &' % (volumn, speed, path))
    raw_input('<enter> to stop audio.')
    os.system("kill -9 `ps aux | grep -v 'grep' | grep afplay | awk '{print $2}'`")


# ===========================================================================
# System query
# ===========================================================================
__process_pid_map = {}


def get_process_status(pid=None, memory_usage=False, memory_shared=False,
                       memory_virtual=False, memory_maps=False,
                       cpu_percent=False, threads=False,
                       status=False, name=False, io_counters=False):
  import psutil
  if pid is None:
    pid = os.getpid()
  if pid in __process_pid_map:
    process = __process_pid_map[pid]
  else:
    process = psutil.Process(pid)
    __process_pid_map[pid] = process

  if status:
    return process.status()
  if name:
    return process.name()
  if io_counters:
    return process.io_counters()
  if memory_usage:
    return process.memory_info().rss / float(2 ** 20)
  if memory_shared:
    return process.memory_info().shared / float(2 ** 20)
  if memory_virtual:
    return process.memory_info().vms / float(2 ** 20)
  if memory_maps:
    return {i[0]: i[1] / float(2 ** 20)
            for i in process.memory_maps()}
  if cpu_percent:
    # first time call always return 0
    process.cpu_percent(None)
    return process.cpu_percent(None)
  if threads:
    return {i.id: (i.user_time, i.system_time) for i in process.threads()}


def get_system_status(memory_total=False, memory_total_actual=False,
                      memory_total_usage=False, memory_total_free=False,
                      all_pids=False, swap_memory=False, pid=False):
  """
  Parameters
  ----------
  threads: bool
      return dict {id: (user_time, system_time)}
  memory_maps: bool
      return dict {path: rss}

  Note
  ----
  All memory is returned in `MiB`
  To calculate memory_percent:
      get_system_status(memory_usage=True) / get_system_status(memory_total=True) * 100
  """
  import psutil
  # ====== general system query ====== #
  if memory_total:
    return psutil.virtual_memory().total / float(2 ** 20)
  if memory_total_actual:
    return psutil.virtual_memory().available / float(2 ** 20)
  if memory_total_usage:
    return psutil.virtual_memory().used / float(2 ** 20)
  if memory_total_free:
    return psutil.virtual_memory().free / float(2 ** 20)
  if swap_memory:
    tmp = psutil.swap_memory()
    tmp.total /= float(2 ** 20)
    tmp.used /= float(2 ** 20)
    tmp.free /= float(2 ** 20)
    tmp.sin /= float(2**20)
    tmp.sout /= float(2**20)
    return tmp
  if all_pids:
    return psutil.pids()
  if pid:
    return os.getpid()



from __future__ import absolute_import, division, print_function

import numpy as np
import tensorflow as tf
from tensorflow_probability.python.bijectors.exp import Exp
from tensorflow_probability.python.distributions import (
    LogNormal, TransformedDistribution, Uniform)
from tensorflow_probability.python.internal import dtype_util

__all__ = [
    "LogUniform",
]


class LogUniform(TransformedDistribution):
  """The log-uniform distribution (i.e. the logarithm of the
  samples from this distribution are Uniform) """

  def __init__(self,
               low=0.,
               high=1.,
               validate_args=False,
               allow_nan_stats=True,
               name="LogUniform"):
    """Construct a log-normal distribution.

    The LogNormal distribution models positive-valued random variables
    whose logarithm is normally distributed with mean `loc` and
    standard deviation `scale`. It is constructed as the exponential
    transformation of a Normal distribution.

    Args:
      low: Floating point tensor, lower boundary of the output interval. Must
        have `low < high`.
      high: Floating point tensor, upper boundary of the output interval. Must
        have `low < high`.
      validate_args: Python `bool`, default `False`. Whether to validate input
        with asserts. If `validate_args` is `False`, and the inputs are
        invalid, correct behavior is not guaranteed.
      allow_nan_stats: Python `bool`, default `True`. If `False`, raise an
        exception if a statistic (e.g. mean/mode/etc...) is undefined for any
        batch member If `True`, batch members with valid parameters leading to
        undefined statistics will return NaN for this statistic.
      name: The name to give Ops created by the initializer.
    """
    parameters = dict(locals())
    with tf.name_scope(name) as name:
      dtype = dtype_util.common_dtype([low, high], tf.float32)
      super(LogUniform, self).__init__(distribution=Uniform(
          low=tf.convert_to_tensor(value=low, name="low", dtype=dtype),
          high=tf.convert_to_tensor(value=high, name="high", dtype=dtype),
          allow_nan_stats=allow_nan_stats),
                                       bijector=Exp(),
                                       validate_args=validate_args,
                                       parameters=parameters,
                                       name=name)

  @staticmethod
  def _param_shapes(sample_shape):
    return dict(
        zip(("low", "high"),
            ([tf.convert_to_tensor(value=sample_shape, dtype=tf.int32)] * 2)))

  @classmethod
  def _params_event_ndims(cls):
    return dict(low=0, high=0)

  @property
  def low(self):
    """Lower boundary of the output interval."""
    return self.distribution.low

  @property
  def high(self):
    """Upper boundary of the output interval."""
    return self.distribution.high

  def range(self, name="range"):
    """`high - low`."""
    with self._name_scope(name):
      return self.high - self.low

  def _entropy(self):
    raise NotImplementedError

  def _mean(self):
    raise NotImplementedError

  def _variance(self):
    raise NotImplementedError

  def _stddev(self):
    raise NotImplementedError



from tensorflow_probability.python.distributions import *

from odin.bay.distributions.logarizmed import *
from odin.bay.distributions.negative_binomial_disp import NegativeBinomialDisp
from odin.bay.distributions.normal_gamma import NormalGamma
from odin.bay.distributions.quantized import *
from odin.bay.distributions.utils import *
from odin.bay.distributions.zero_inflated import ZeroInflated



from odin.networks.advance_model import AdvanceModel
from odin.networks.distribution_util_layers import *
from odin.networks.math import *
from odin.networks.mixture_density_network import *
from odin.networks.stat_layers import *
from odin.networks.util_layers import *


def register_new_keras_layers(extras=None):
  from tensorflow.python.keras.layers import Layer
  from tensorflow.python.keras.utils.generic_utils import _GLOBAL_CUSTOM_OBJECTS
  globs = dict(globals())
  if extras is not None:
    globs.update(extras)
  for key, val in globs.items():
    if isinstance(val, type) and issubclass(val, Layer):
      _GLOBAL_CUSTOM_OBJECTS[key] = val


register_new_keras_layers()



from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')

import os
os.environ['ODIN'] = 'float32,gpu'

import scipy.io
import numpy as np

import tensorflow as tf
from tensorflow.python.ops import init_ops

from odin import training
from odin.utils import (args_parse, ctext, Progbar, as_tuple_of_shape,
                        crypto, stdio)
from odin import fuel as F, visual as V, nnet as N, backend as K

from utils import prepare_dnn_data, get_model_path, csv2mat
# ===========================================================================
# Configs
# ===========================================================================
args = args_parse([
    ('recipe', 'the name of function defined in feature_recipes.py', None),
    ('-feat', "Acoustic feature", ('mspec', 'mfcc'), 'mspec'),
    ('-batch', "batch size", None, 64),
    ('-epoch', "number of epoch", None, 25),
    ('-l', "audio segmenting length in second", None, 3),
    ('--debug', "enable debug mode", None, False),
    ('--train', "force continue training the saved model", None, False),
])
FEAT = args.feat
TRAIN_MODEL = args.train
DEBUG = bool(args.debug)
(EXP_DIR, MODEL_PATH, LOG_PATH,
 TRAIN_PATH, TEST_PATH) = get_model_path('xvec', args)
stdio(LOG_PATH)
# ===========================================================================
# Create data feeder
# ===========================================================================
(train, valid,
 test_ids, test_dat,
 all_speakers) = prepare_dnn_data(
    recipe=args.recipe, feat=FEAT, utt_length=args.l)
n_speakers = len(all_speakers) + 1
# ===========================================================================
# Create the network
# ===========================================================================
inputs = [K.placeholder(shape=(None,) + shape[1:],
                        dtype='float32',
                        name='input%d' % i)
          for i, shape in enumerate(as_tuple_of_shape(train.shape))]
X = inputs[0]
y = inputs[1]
print("Inputs:", ctext(inputs, 'cyan'))
# ====== the network ====== #
if os.path.exists(MODEL_PATH):
  x_vec = N.deserialize(path=MODEL_PATH, force_restore_vars=True)
else:
  TRAIN_MODEL = True
  with N.args_scope(
      ['TimeDelayedConv', dict(time_pool='none', activation=K.relu)],
      ['Dense', dict(activation=K.linear, b_init=None)],
      ['BatchNorm', dict(activation=K.relu)]
  ):
    x_vec = N.Sequence([
        N.Dropout(level=0.3),

        N.TimeDelayedConv(n_new_features=512, n_time_context=5),
        N.TimeDelayedConv(n_new_features=512, n_time_context=5),
        N.TimeDelayedConv(n_new_features=512, n_time_context=7),

        N.Dense(512), N.BatchNorm(),
        N.Dense(1500), N.BatchNorm(),

        N.StatsPool(axes=1, output_mode='concat'),
        N.Flatten(outdim=2),

        N.Dense(512, name="LatentOutput"), N.BatchNorm(),
        N.Dense(512), N.BatchNorm(),

        N.Dense(n_speakers, activation=K.linear,
                b_init=init_ops.constant_initializer(value=0))
    ], debug=1)
# ====== create outputs ====== #
y_logit = x_vec(X)
y_proba = tf.nn.softmax(y_logit)
z = K.ComputationGraph(y_proba).get(roles=N.Dense, scope='LatentOutput',
                                    beginning_scope=False)[0]
print('Latent space:', ctext(z, 'cyan'))
# ====== create loss ====== #
ce = tf.losses.softmax_cross_entropy(onehot_labels=y, logits=y_logit)
acc = K.metrics.categorical_accuracy(y_true=y, y_pred=y_proba)
# ====== params and optimizing ====== #
updates = K.optimizers.Adam(lr=0.0001, name='XAdam').minimize(
    loss=ce,
    roles=[K.role.TrainableParameter],
    exclude_roles=[K.role.InitialState],
    verbose=True)
K.initialize_all_variables()
# # ====== Functions ====== #
print('Building training functions ...')
f_train = K.function(inputs, [ce, acc], updates=updates,
                     training=True)
print('Building testing functions ...')
f_score = K.function(inputs, [ce, acc],
                    training=False)
# Latent spaces
f_z = K.function(inputs=X, outputs=z, training=False)
# ===========================================================================
# Create trainer
# ===========================================================================
if TRAIN_MODEL:
  print('Start training ...')
  task = training.MainLoop(batch_size=args.batch, seed=1234, shuffle_level=2,
                           allow_rollback=True)
  task.set_checkpoint(MODEL_PATH, x_vec)
  task.set_callbacks([
      training.NaNDetector(),
      training.EarlyStopGeneralizationLoss('valid', ce,
                                           threshold=5, patience=5)
  ])
  task.set_train_task(func=f_train, data=train,
                      epoch=args.epoch, name='train')
  task.set_valid_task(func=f_score, data=valid,
                      freq=training.Timer(percentage=0.8),
                      name='valid')
  task.run()
# ===========================================================================
# Saving the test data
# CSV separated by tab
# ===========================================================================
sep = '\t'
prog = Progbar(target=len(test_ids) + len(train) + len(valid),
               print_summary=True, print_report=True,
               name="Extracting x-vector")
with open(TRAIN_PATH, 'w') as f_train, open(TEST_PATH, 'w') as f_test:
  # ====== save training set ====== #
  for name, idx, X, y in train.set_batch(batch_size=8000,
                                         batch_mode='file', seed=None):
    assert idx == 0
    y = np.argmax(y, axis=-1)
    assert len(set(y)) == 1
    y = y[0]
    z = np.mean(f_z(X), axis=0, keepdims=False).astype('float32')
    f_train.write(sep.join([str(y)] + [str(i) for i in z]) + '\n')
    prog.add(X.shape[0])
  # ====== save validation set ====== #
  for name, idx, X, y in valid.set_batch(batch_size=8000,
                                         batch_mode='file', seed=None):
    assert idx == 0
    y = np.argmax(y, axis=-1)
    assert len(set(y)) == 1
    y = y[0]
    z = np.mean(f_z(X), axis=0, keepdims=False).astype('float32')
    f_train.write(sep.join([str(y)] + [str(i) for i in z]) + '\n')
    prog.add(X.shape[0])
  # ====== save test set ====== #
  for name, (start, end) in sorted(test_ids.items(),
                                   key=lambda x: x[0]):
    y = test_dat[start:end]
    z = np.mean(f_z(y), axis=0, keepdims=False).astype('float32')
    f_test.write(sep.join([name] + [str(i) for i in z]) + '\n')
    prog.add(1)
# convert everything to matlab format
csv2mat(exp_dir=EXP_DIR)
# ===========================================================================
# Evaluate and save the log
# ===========================================================================
np.random.seed(87654321)
shape = inputs[0].shape
X = np.random.rand(64, shape[1].value, shape[2].value).astype('float32')
Z = f_z(X)
# ====== make sure model has the same identity ====== #
print(Z.shape, Z.sum(), (Z**2).sum(), Z.std())
print(ctext(crypto.md5_checksum(Z), 'cyan'))



from __future__ import absolute_import, division, print_function

import numpy as np
from tensorflow_probability.python import distributions as tfd
from tensorflow_probability.python.layers import distribution_layer as tfl

from odin.bay import distribution_layers as obl
from odin.bay import distributions as obd
from odin.utils.python_utils import multikeysdict

# mapping from alias to
_dist_mapping = multikeysdict({
    ('bern', 'bernoulli'): (obl.BernoulliLayer, tfd.Bernoulli),
    ('zibernoulli', 'zeroinflatedbernoulli'):
        (obl.ZIBernoulliLayer, tfd.Bernoulli),
    ('normal', 'gaussian'): (obl.NormalLayer, tfd.Normal),
    'lognormal': (obl.LogNormalLayer, tfd.LogNormal),
    ('nb', 'negativebinomial'):
        (obl.NegativeBinomialLayer, tfd.NegativeBinomial),
    ('zinb', 'zinegativebinomial', 'zeroinflatednegativebinomial'):
        (obl.ZINegativeBinomialLayer, tfd.NegativeBinomial),
    ('nbd', 'negativebinomialdisp'):
        (obl.NegativeBinomialDispLayer, obd.NegativeBinomialDisp),
    ('zinbd', 'zinegativebinomialdisp', 'zeroinflatednegativebinomialdisp'):
        (obl.ZINegativeBinomialDispLayer, obd.NegativeBinomialDisp),
    ('pois', 'poisson'): (obl.PoissonLayer, tfd.Poisson),
    ('zipois', 'zipoisson', 'zeroinflatedpoisson'):
        (obl.ZIPoissonLayer, tfd.Poisson),
    'dirichlet': (obl.DirichletLayer, tfd.Dirichlet),
    'onehot': (obl.OneHotCategoricalLayer, tfd.OneHotCategorical),
    'deterministic': (obl.DeterministicLayer, tfd.Deterministic),
    'vdeterministic': (obl.VectorDeterministicLayer, tfd.VectorDeterministic),
    # 'beta': (obl),
})


def parse_distribution(alias):
  """
  Return
  ------
  layer : `tensorflow_probability.python.layers.DistributionLambda`
  dist : `tensorflow_probability.python.distributions.Distribution`
  """
  alias = str(alias).lower()
  if alias not in _dist_mapping:
    raise ValueError("Cannot find distribution with alias: '%s', "
                     "all available distributions: %s" %
                     (alias, ', '.join(list(_dist_mapping.keys()))))
  layer, dist = _dist_mapping[alias]
  return layer, dist



from __future__ import absolute_import, division, print_function

import numpy as np
import tensorflow.compat.v2 as tf
from tensorflow_probability.python.bijectors import exp as exp_bijector
from tensorflow_probability.python.distributions import (
    NegativeBinomial, Normal, QuantizedDistribution, TransformedDistribution,
    Uniform)
from tensorflow_probability.python.internal import dtype_util

__all__ = ["qUniform", "qNormal"]


class qNormal(QuantizedDistribution):

  def __init__(self,
               loc=0.,
               scale=1.,
               min_value=None,
               max_value=None,
               validate_args=False,
               allow_nan_stats=True,
               name="qNormal"):
    super(qNormal,
          self).__init__(distribution=Normal(loc=loc,
                                             scale=scale,
                                             validate_args=validate_args,
                                             allow_nan_stats=allow_nan_stats),
                         low=min_value,
                         high=max_value,
                         name=name)


class qUniform(QuantizedDistribution):

  def __init__(self,
               low=0.,
               high=1.,
               min_value=None,
               max_value=None,
               validate_args=False,
               allow_nan_stats=True,
               name="qUniform"):
    super(qUniform,
          self).__init__(distribution=Uniform(low=low,
                                              high=high,
                                              validate_args=validate_args,
                                              allow_nan_stats=allow_nan_stats),
                         low=min_value,
                         high=max_value,
                         name=name)



"""The Alternative parameterization for Negative Binomial distribution class."""
from __future__ import absolute_import, division, print_function

import tensorflow as tf
from tensorflow_probability.python.distributions import distribution
from tensorflow_probability.python.internal import (assert_util,
                                                    distribution_util,
                                                    dtype_util,
                                                    reparameterization)
from tensorflow_probability.python.util.seed_stream import SeedStream


class NegativeBinomialDisp(distribution.Distribution):
  """Alternate parameterization for NegativeBinomial distribution using
  mean and dispersion as often mentioned in infobiomatic literatures.

  The NegativeBinomial distribution is related to the experiment of performing
  Bernoulli trials in sequence. Given a Bernoulli trial with probability `p` of
  success, the NegativeBinomial distribution represents the distribution over
  the number of successes `s` that occur until we observe `f` failures.

  The probability mass function (pmf) is,

  ```none
  pmf(s; f, p) = p**s (1 - p)**f / Z
  Z = s! (f - 1)! / (s + f - 1)!
  ```

  where:
  * `total_count = f`,
  * `probs = p`,
  * `Z` is the normalizaing constant, and,
  * `n!` is the factorial of `n`.
  """

  def __init__(self,
               loc,
               disp,
               validate_args=False,
               allow_nan_stats=True,
               name="NegativeBinomialDisp"):
    """Construct NegativeBinomial distributions.

    Args:
      loc: Non-negative floating-point `Tensor` with shape
        broadcastable to `[B1,..., Bb]` with `b >= 0` and the same dtype as
        `probs` or `logits`. Defines this as a batch of `N1 x ... x Nm`
        different Negative Binomial distributions. In practice, this represents
        the number of negative Bernoulli trials to stop at (the `total_count`
        of failures), but this is still a valid distribution when
        `total_count` is a non-integer.
      disp: Non-negative floating-point `Tensor` with shape broadcastable to
        `[B1, ..., Bb]` where `b >= 0` indicates the number of batch dimensions.
        Each entry represents logits for the probability of success for
        independent Negative Binomial distributions and must be in the open
        interval `(-inf, inf)`. Only one of `logits` or `probs` should be
        specified. (also known as `theta`)
      validate_args: Python `bool`, default `False`. When `True` distribution
        parameters are checked for validity despite possibly degrading runtime
        performance. When `False` invalid inputs may silently render incorrect
        outputs.
      allow_nan_stats: Python `bool`, default `True`. When `True`, statistics
        (e.g., mean, mode, variance) use the value "`NaN`" to indicate the
        result is undefined. When `False`, an exception is raised if one or
        more of the statistic's batch members are undefined.
      name: Python `str` name prefixed to Ops created by this class.
    """

    parameters = dict(locals())
    with tf.name_scope(name) as name:
      dtype = dtype_util.common_dtype([loc, disp], dtype_hint=tf.float32)
      loc = tf.convert_to_tensor(value=loc, name="loc", dtype=dtype)
      disp = tf.convert_to_tensor(value=disp, name="disp", dtype=dtype)
      with tf.control_dependencies(
          [assert_util.assert_positive(loc),
           assert_util.assert_positive(disp)] if validate_args else []):
        self._loc = tf.identity(loc, name="loc")
        self._disp = tf.identity(disp, name="disp")
    super(NegativeBinomialDisp, self).__init__(
        dtype=self._loc.dtype,
        reparameterization_type=reparameterization.NOT_REPARAMETERIZED,
        validate_args=validate_args,
        allow_nan_stats=allow_nan_stats,
        parameters=parameters,
        graph_parents=[self._loc, self._disp],
        name=name)

  @classmethod
  def _params_event_ndims(cls):
    return dict(loc=0, disp=0)

  @property
  def loc(self):
    """Mean."""
    return self._loc

  @property
  def disp(self):
    """Dispersion"""
    return self._disp

  def _batch_shape_tensor(self):
    return tf.broadcast_dynamic_shape(tf.shape(input=self.loc),
                                      tf.shape(input=self.disp))

  def _batch_shape(self):
    return tf.broadcast_static_shape(self.loc.shape, self.disp.shape)

  def _event_shape_tensor(self):
    return tf.constant([], dtype=tf.int32)

  def _event_shape(self):
    return tf.TensorShape([])

  def _sample_n(self, n, seed=None):
    # Here we use the fact that if:
    # p = loc / (loc + disp)
    # r = disp
    # concentration = r
    # rate = (1-p)/p
    # lam ~ Gamma(concentration, rate
    # then X ~ Poisson(lam) is Negative Binomially distributed.
    stream = SeedStream(seed, salt="NegativeBinomialDisp")
    p = self.loc / (self.loc + self.disp)
    r = self.disp
    concentration = r
    rate = (1 - p) / p
    rate = tf.random.gamma(shape=[n],
                           alpha=concentration,
                           beta=rate,
                           dtype=self.dtype,
                           seed=stream())
    return tf.random.poisson(lam=rate,
                             shape=[],
                             dtype=self.dtype,
                             seed=stream())

  def _cdf(self, x):
    if self.validate_args:
      x = distribution_util.embed_check_nonnegative_integer_form(x)
    raise NotImplementedError()

  def _log_prob(self, x):
    return self._log_unnormalized_prob(x) - self._log_normalization(x)

  def _log_unnormalized_prob(self, x, eps=1e-8):
    if self.validate_args:
      x = distribution_util.embed_check_nonnegative_integer_form(x)
    eps = tf.cast(eps, dtype=self.dtype)
    log_loc_disp_eps = tf.math.log(self.disp + self.loc + eps)
    return self.disp * (tf.math.log(self.disp + eps) - log_loc_disp_eps) \
            + x * (tf.math.log(self.loc + eps) - log_loc_disp_eps)

  def _log_normalization(self, x):
    if self.validate_args:
      x = distribution_util.embed_check_nonnegative_integer_form(x)
    return tf.math.lgamma(self.disp) \
      + tf.math.lgamma(x + 1) \
        - tf.math.lgamma(x + self.disp)

  def _mean(self):
    return self._loc

  def _mode(self):
    raise NotImplementedError()

  def _variance(self):
    mean = self._mean()
    return mean + tf.square(mean) / self.disp



from __future__ import print_function, division, absolute_import

import numpy as np
from odin.utils.mpi import MPI, cpu_count
from odin.utils import wprint
from odin.utils.crypto import md5_checksum

_cached_values = {}

# ===========================================================================
# auto-select best TSNE
# ===========================================================================
def _create_key(kwargs, md5):
  key = dict(kwargs)
  del key['verbose']
  key['md5'] = md5
  return str(list(sorted(key.items(), key=lambda x: x[0])))

def fast_tsne(*X, n_components=2, n_samples=None, perplexity=30.0,
              early_exaggeration=8.0, learning_rate=200.0, n_iter=1000,
              n_iter_without_progress=300, min_grad_norm=1e-7,
              metric="euclidean", init="random", verbose=0,
              random_state=1234, method='barnes_hut', angle=0.5,
              n_jobs=4):
  """
  Parameters
  ----------
  n_components : int, optional (default: 2)
      Dimension of the embedded space.

  n_samples : {int, None}
      if given, downsampling the data to given number of sample

  perplexity : float, optional (default: 30)
      The perplexity is related to the number of nearest neighbors that
      is used in other manifold learning algorithms. Larger datasets
      usually require a larger perplexity. Consider selecting a value
      between 5 and 50. The choice is not extremely critical since t-SNE
      is quite insensitive to this parameter.

  early_exaggeration : float, optional (default: 8.0)
      Controls how tight natural clusters in the original space are in
      the embedded space and how much space will be between them. For
      larger values, the space between natural clusters will be larger
      in the embedded space. Again, the choice of this parameter is not
      very critical. If the cost function increases during initial
      optimization, the early exaggeration factor or the learning rate
      might be too high.

  learning_rate : float, optional (default: 200.0)
      The learning rate for t-SNE is usually in the range [10.0, 1000.0]. If
      the learning rate is too high, the data may look like a 'ball' with any
      point approximately equidistant from its nearest neighbours. If the
      learning rate is too low, most points may look compressed in a dense
      cloud with few outliers. If the cost function gets stuck in a bad local
      minimum increasing the learning rate may help.

  n_iter : int, optional (default: 1000)
      Maximum number of iterations for the optimization. Should be at
      least 250.

  n_iter_without_progress : int, optional (default: 300)
      Maximum number of iterations without progress before we abort the
      optimization, used after 250 initial iterations with early
      exaggeration. Note that progress is only checked every 50 iterations so
      this value is rounded to the next multiple of 50.

  min_grad_norm : float, optional (default: 1e-7)
      If the gradient norm is below this threshold, the optimization will
      be stopped.

  metric : string or callable, optional
      The metric to use when calculating distance between instances in a
      feature array. If metric is a string, it must be one of the options
      allowed by scipy.spatial.distance.pdist for its metric parameter, or
      a metric listed in pairwise.PAIRWISE_DISTANCE_FUNCTIONS.
      If metric is "precomputed", X is assumed to be a distance matrix.
      Alternatively, if metric is a callable function, it is called on each
      pair of instances (rows) and the resulting value recorded. The callable
      should take two arrays from X as input and return a value indicating
      the distance between them. The default is "euclidean" which is
      interpreted as squared euclidean distance.

  init : string or numpy array, optional (default: "random")
      Initialization of embedding. Possible options are 'random', 'pca',
      and a numpy array of shape (n_samples, n_components).
      PCA initialization cannot be used with precomputed distances and is
      usually more globally stable than random initialization.

  verbose : int, optional (default: 0)
      Verbosity level.

  random_state : int, RandomState instance or None, optional (default: None)
      If int, random_state is the seed used by the random number generator;
      If RandomState instance, random_state is the random number generator;
      If None, the random number generator is the RandomState instance used
      by `np.random`.  Note that different initializations might result in
      different local minima of the cost function.

  method : string (default: 'barnes_hut')
      By default the gradient calculation algorithm uses Barnes-Hut
      approximation running in O(NlogN) time. method='exact'
      will run on the slower, but exact, algorithm in O(N^2) time. The
      exact algorithm should be used when nearest-neighbor errors need
      to be better than 3%. However, the exact method cannot scale to
      millions of examples.

  angle : float (default: 0.5)
      Only used if method='barnes_hut'
      This is the trade-off between speed and accuracy for Barnes-Hut T-SNE.
      'angle' is the angular size (referred to as theta in [3]) of a distant
      node as measured from a point. If this size is below 'angle' then it is
      used as a summary node of all points contained within it.
      This method is not very sensitive to changes in this parameter
      in the range of 0.2 - 0.8. Angle less than 0.2 has quickly increasing
      computation time and angle greater 0.8 has quickly increasing error.
  """
  assert len(X) > 0, "No input is given!"
  if isinstance(X[0], (tuple, list)):
    X = X[0]
  if not all(isinstance(x, np.ndarray) for x in X):
    raise ValueError("`X` can only be list of numpy.ndarray or numpy.ndarray")
  # ====== kwarg for creating T-SNE class ====== #
  kwargs = dict(locals())
  del kwargs['X']
  n_samples = kwargs.pop('n_samples', None)
  # ====== downsampling ====== #
  if n_samples is not None:
    n_samples = int(n_samples)
    assert n_samples > 0
    new_X = []
    rand = random_state if isinstance(random_state, np.random.RandomState) else \
    np.random.RandomState(seed=random_state)
    for x in X:
      if x.shape[0] > n_samples:
        ids = rand.permutation(x.shape[0])[:n_samples]
        x = x[ids]
      new_X.append(x)
    X = new_X
  # ====== import proper T-SNE ====== #
  tsne_version = None
  try:
    from tsnecuda import TSNE
    from tsnecuda.NaiveTSNE import NaiveTSNE as _exact_TSNE
    tsne_version = 'cuda'
  except ImportError:
    # wprint("Install CUDA-TSNE from `https://github.com/CannyLab/tsne-cuda` "
    #        "for significant speed up.")
    try:
      from MulticoreTSNE import MulticoreTSNE as TSNE
      tsne_version = 'multicore'
    except ImportError:
      wprint("Install MulticoreTSNE from `pip install git+https://github.com/DmitryUlyanov/Multicore-TSNE.git`"
             ' to accelerate the T-SNE on multiple CPU cores.')
      try:
        from sklearn.manifold import TSNE
        tsne_version = 'sklearn'
      except Exception as e:
        raise e
  # ====== modify kwargs ====== #
  if tsne_version == 'cuda':
    kwargs['random_seed'] = kwargs['random_state']
    kwargs['theta'] = angle
    if method == 'exact':
      TSNE = _exact_TSNE
      del kwargs['theta']
    del kwargs['random_state']
    del kwargs['n_jobs']
    del kwargs['angle']
    del kwargs['method']
  elif tsne_version == 'multicore':
    pass
  else:
    del kwargs['n_jobs']
  # ====== getting cached values ====== #
  results = []
  X_new = []
  for i, x in enumerate(X):
    md5 = md5_checksum(x)
    key = _create_key(kwargs, md5)
    if key in _cached_values:
      results.append((i, _cached_values[key]))
    else:
      X_new.append((i, md5, x))

  # ====== perform T-SNE ====== #
  def apply_tsne(j):
    idx, md5, x = j
    tsne = TSNE(**kwargs)
    return (idx, md5, tsne.fit_transform(x))
  # only 1 X, no need for MPI
  if len(X_new) == 1:
    idx, md5, x = apply_tsne(X_new[0])
    results.append((idx, x))
    _cached_values[_create_key(kwargs, md5)] = x
  else:
    mpi = MPI(jobs=X_new, func=apply_tsne, batch=1,
              ncpu=min(len(X_new), cpu_count() - 1))
    for idx, md5, x in mpi:
      results.append((idx, x))
      _cached_values[_create_key(kwargs, md5)] = x
  # ====== return and clean ====== #
  results = sorted(results, key=lambda a: a[0])
  results = [r[1] for r in results]
  return results[0] if len(results) == 1 else results
